__all__ = ('create_payload', )

import base64, binascii, mimetypes as mime_types, os, re, uuid
from io import BufferedRandom, BufferedReader, BytesIO, IOBase, StringIO, TextIOBase
from urllib.parse import urlencode as url_encode

from ..core import AsyncIO
from ..utils import IgnoreCaseMultiValueDictionary, to_json

from .compressors import BROTLI_COMPRESSOR, ZLIB_COMPRESSOR, ZLIB_MAX_WBITS
from .exceptions import ContentEncodingError
from .header_building_and_parsing import CHARS, TOKENS, build_content_disposition_header
from .headers import CONTENT_DISPOSITION, CONTENT_ENCODING, CONTENT_LENGTH, CONTENT_TRANSFER_ENCODING, CONTENT_TYPE
from .mime_type import MimeType
from .quoting import unquote


BIG_CHUNK_LIMIT = 1 << 16
DEFAULT_CONTENT_TYPE = 'application/octet-stream'

VALID_TCHAR_RP = re.compile(br'\A[!#$%&\'*+\-.^_`|~\w]+\Z')
INVALID_QDTEXT_CHAR_RP = re.compile(br'[\x00-\x08\x0A-\x1F\x7F]')

def create_payload(data, kwargs):
    """
    Creates a new ``PayloadBase`` with the given parameters.
    
    Parameters
    ----------
    data : (`list` of ``PayloadBase``), ``BodyPartReader``, `bytes`, `bytearray`, `memoryview`, `str`, \
            `BytesIO`, `StringIO`, `TextIOBase`, `BufferedReader`, `BufferedRandom`, `IOBase`, ``AsyncIO``, \
            `async-iterable`
    kwargs : `dict` of (`str`, (`None`, `str`)) items
        Keyword parameters for the payload.
    
    Returns
    -------
    payload : ``PayloadBase``
    
    Raises
    ------
    LookupError
        `payload` is not serializable.
    """
    data_type = data.__class__
    #if issubclass(data_type, BodyPartReader):
    #    type_ = BodyPartReaderPayload
    if issubclass(data_type, (bytes, bytearray, memoryview)):
        type_ = BytesPayload
    elif issubclass(data_type, str):
        type_ = StringPayload
    elif issubclass(data_type, BytesIO):
        type_ = BytesIOPayload
    elif issubclass(data_type, StringIO):
        type_ = StringIOPayload
    elif issubclass(data_type, TextIOBase):
        type_ = TextIOPayload
    elif issubclass(data_type, (BufferedReader, BufferedRandom)):
        type_ = BufferedReaderPayload
    elif issubclass(data_type, IOBase):
        type_ = IOBasePayload
    elif issubclass(data_type, AsyncIO):
        type_ = AsyncIOPayload
    elif hasattr(data_type, '__aiter__'):
        type_ = AsyncIterablePayload
    else:
        raise LookupError(data_type)
    
    return type_(data, kwargs)


class PayloadBase:
    """
    Base class for payloads.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data : (`list` of ``PayloadBase``), ``BodyPartReader``, `bytes`, `bytearray`, `memoryview`, `str`, \
            `BytesIO`, `StringIO`, `TextIOBase`, `BufferedReader`, `BufferedRandom`, `IOBase`, ``AsyncIO``, \
            `async-iterable`
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    __slots__ = ('content_type', 'data', 'encoding', 'filename', 'headers', 'size', )
    
    def __init__(self, data, kwargs):
        """
        Creates a new ``PayloadBase``.
        
        Parameters
        ----------
        data : (`list` of ``PayloadBase``), ``BodyPartReader``, `bytes`, `bytearray`, `memoryview`, `str`, \
                `BytesIO`, `StringIO`, `TextIOBase`, `BufferedReader`, `BufferedRandom`, `IOBase`, ``AsyncIO``, \
                `async-iterable`
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword parameters.
        """
        self.data = data
        self.encoding = kwargs.get('encoding', None)
        self.filename = filename = kwargs.get('filename', None)
        self.size = None
        
        headers = IgnoreCaseMultiValueDictionary()
        
        content_type = kwargs.get('content_type', None)
        if content_type is None:
            if filename is None:
                content_type = DEFAULT_CONTENT_TYPE
            else:
                mime = mime_types.guess_type(self.filename)[0]
                if mime is None:
                    content_type = DEFAULT_CONTENT_TYPE
                else:
                    content_type = mime
        
        headers[CONTENT_TYPE] = content_type
        
        headers_parameter = kwargs.get('headers', None)
        if (headers_parameter is not None) or  headers_parameter:
            headers.extend(headers_parameter)
        
        self.headers = headers
        self.content_type = content_type
    
    def set_content_disposition(self, disposition_type, parameters, quote_fields):
        """
        Sets content disposition header to the payload.
        
        Parameters
        ----------
        disposition_type : `str`
            The disposition's type.
        parameters : `dict` of (`str`, `str`) items
            Additional parameters.
        quote_fields : `bool`
            Whether field values should be quoted.
        """
        headers = self.headers
        headers.pop_all(CONTENT_DISPOSITION, None)
        headers[CONTENT_DISPOSITION] = build_content_disposition_header(disposition_type, parameters, quote_fields)
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        pass

class BytesPayload(PayloadBase):
    """
    Payload class for `bytes-like` objects.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data : `bytes`, `bytearray`, `memoryview`
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    
    def __init__(self, data, kwargs):
        """
        Creates a new ``BytesPayload``.
        
        Parameters
        ----------
        data : `bytes`, `bytearray`, `memoryview`
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword parameters.
        """
        kwargs.setdefault('content_type', DEFAULT_CONTENT_TYPE)
        
        PayloadBase.__init__(self, data, kwargs)
        
        self.size = len(data)
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        await writer.write(self.data)


class StringPayload(BytesPayload):
    """
    Payload class for `bytes-like` objects.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data : `bytes`
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    
    def __init__(self, data, kwargs):
        """
        Creates a new ``StringPayload``.
        
        Parameters
        ----------
        data : `str`
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword parameters.
        """
        encoding = kwargs.get('encoding', None)
        content_type = kwargs.get('content_type', None)
        if encoding is None:
            if content_type is None:
                encoding = 'utf-8'
                content_type = 'text/plain; charset=utf-8'
                
                kwargs['content_type'] = content_type
            else:
                mime_type = MimeType(content_type)
                encoding = mime_type.parameters.get('charset', 'utf-8')
            
            kwargs['encoding'] = encoding
        
        else:
            if content_type is None:
                content_type = f'text/plain; charset={encoding}'
                kwargs['content_type'] = content_type
        
        data = data.encode(encoding)
        
        BytesPayload.__init__(self, data, kwargs)


class StringIOPayload(StringPayload):
    """
    Payload class for string io objects.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data : `bytes`
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    
    def __init__(self, data, kwargs):
        """
        Creates a new ``StringIOPayload``.
        
        Parameters
        ----------
        data : `StringIO`
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword parameters.
        """
        data = data.read()
        StringPayload.__init__(self, data, kwargs)


class IOBasePayload(PayloadBase):
    """
    Payload class for `IOBase`-s.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data : `IOBase`
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    def __init__(self, data, kwargs):
        """
        Creates a new ``IOBasePayload``.
        
        Parameters
        ----------
        data : `IOBase`
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword parameters.
        """
        if 'filename' not in kwargs:
            kwargs['filename'] = getattr(data, 'name', None)
        
        PayloadBase.__init__(self, data, kwargs)
        
        try:
            disposition = kwargs['disposition']
        except KeyError:
            disposition = 'attachment'
        
        if (disposition is not None):
            filename = self.filename
            if (filename is not None):
                self.set_content_disposition(disposition, {'filename': filename}, True)
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        data = self.data
        try:
            while True:
                chunk = data.read(BIG_CHUNK_LIMIT)
                if chunk:
                    await writer.write(chunk)
                else:
                    break
        finally:
            data.close()


class TextIOPayload(IOBasePayload):
    """
    Payload class for `TextIOBase`-s.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data : `TextIOBase`
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    
    def __init__(self, data, kwargs):
        """
        Creates a new ``TextIOPayload``.
        
        Parameters
        ----------
        data : `TextIOBase`
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword parameters.
        """
        encoding = kwargs.get('encoding', None)
        content_type = kwargs.get('content_type', None)
        if encoding is None:
            if content_type is None:
                encoding = 'utf-8'
                content_type = 'text/plain; charset=utf-8'
                
                kwargs['content_type'] = content_type
            else:
                mime_type = MimeType(content_type)
                encoding = mime_type.parameters.get('charset', 'utf-8')
            
            kwargs['encoding'] = encoding
        else:
            if content_type is None:
                content_type = f'text/plain; charset={encoding}'
                kwargs['content_type'] = content_type
        
        IOBasePayload.__init__(self, data, kwargs)
        
        try:
            size = os.fstat(data.fileno()).st_size - data.tell()
        except OSError:
            # `data.fileno()` is not supported. Example: `io.BufferedReader(io.BytesIO(b'data'))`
            size = None
        
        self.size = size
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        data = self.data
        try:
            while True:
                chunk = data.read(BIG_CHUNK_LIMIT)
                if chunk:
                    await writer.write(chunk.encode(self.encoding))
                else:
                    break
        finally:
            data.close()


class BytesIOPayload(IOBasePayload):
    """
    Payload class for `BytesIO`-s.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data : `BytesIO`
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    def __init__(self, data, kwargs):
        """
        Creates a new ``BytesIOPayload``.
        
        Parameters
        ----------
        data : `BytesIO`
            The payload's data.
        kwargs : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
            Additional keyword parameters.
        """
        IOBasePayload.__init__(self, data, kwargs)
        
        position = data.tell()
        end = data.seek(0, os.SEEK_END)
        data.seek(position)
        self.size = end - position


class BufferedReaderPayload(IOBasePayload):
    """
    Payload class for `BufferedReader` and for `BufferedRandom`-s.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data : `BufferedReader`, `BufferedRandom`
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    
    def __init__(self, data, kwargs):
        """
        Creates a new ``BufferedReaderPayload``.
        
        Parameters
        ----------
        data : `BufferedReader`, `BufferedRandom`
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword parameters.
        """
        IOBasePayload.__init__(self, data, kwargs)
        try:
            size = os.fstat(data.fileno()).st_size - data.tell()
        except OSError:
            # `data.fileno()` is not supported. Example: `io.BufferedReader(io.BytesIO(b'data'))`
            size = None
        
        self.size = size


class JsonPayload(BytesPayload):
    """
    Payload class for `json` data.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data :`bytes`
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    
    def __init__(self, data, kwargs):
        """
        Creates a new ``AsyncIterablePayload``.
        
        Parameters
        ----------
        data : `None`, `str`, `int`, `float`, `list` of repeat, `dict` of (`str`, repeat) items
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword parameters.
        """
        encoding = kwargs.get('encoding', None)
        if (encoding is None):
            kwargs['encoding'] = encoding = 'utf-8'
        
        data = to_json(data).encode(encoding)
        
        kwargs.setdefault('content_type', 'application/json')
        BytesPayload.__init__(self, data, kwargs)


class AsyncIterablePayload(PayloadBase):
    """
    Payload class for `async-iterable`-s.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data :`async-iterable`
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    __slots__ = ('_iterator',)
    
    def __init__(self, data, kwargs):
        """
        Creates a new ``AsyncIterablePayload``.
        
        Parameters
        ----------
        data : `async-iterable`
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword parameters.
        """
        kwargs.setdefault('content_type', DEFAULT_CONTENT_TYPE)
        
        PayloadBase.__init__(self, data, kwargs)
        self._iterator = data.__class__.__aiter__(data)
    
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        iterator = self._iterator
        anext = iterator.__class__.__anext__
        try:
            while True:
                chunk = await anext(iterator)
                await writer.write(chunk)
        except StopAsyncIteration:
            self._iterator = None


class AsyncIOPayload(IOBasePayload):
    """
    Payload class for ``AsyncIO``-s.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data :`async-iterable`
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        data = self.data
        try:
            while True:
                chunk = await data.read(BIG_CHUNK_LIMIT)
                await writer.write(chunk)
                if len(chunk) < BIG_CHUNK_LIMIT:
                    break
        finally:
            data.close()


class BodyPartReaderPayload(PayloadBase):
    """
    Payload class for ``BodyPartReader``.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data :``BodyPartReader``
        The payload itself.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    """
    def __init__(self, data, kwargs):
        """
        Creates a new ``BodyPartReaderPayload``.
        
        Parameters
        ----------
        data : ``BodyPartReader``
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword parameters.
        """
        PayloadBase.__init__(self, data, kwargs)
        
        parameters = {}
        name = data.name
        if (name is not None):
            parameters['name'] = name
        
        filename = data.filename
        if (filename is not None):
            parameters['filename'] = filename
        
        if parameters:
            self.set_content_disposition('attachment', parameters, True)
    
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        field = self.data
        while True:
            chunk = await field.read_chunk(size=65536)
            if chunk:
                await writer.write(field.decode(chunk))
            else:
                break


def _is_continuous_parameter(string):
    """
    Returns whether the content disposition parameter is continuous.
    
    Parameters
    ----------
    string : `str`
        The string to check.
    
    Returns
    -------
    is_continuous_parameter : `bool`
    """
    pos = string.find('*')
    if pos == -1:
        return False
    
    pos += 1
    
    if string.endswith('*'):
        substring = string[pos:-1]
    else:
        substring = string[pos:]
    
    return substring.isdigit()


def _is_token(string):
    """
    Returns whether the given string can be a token of a content disposition parameter.
    
    Parameters
    ----------
    string : `str`
        The string to check.

    Returns
    -------
    is_token : `bool`
    """
    if not string:
        return False
    
    if TOKENS < set(string):
        return False
    
    return True


def _is_quoted(string):
    """
    Returns whether the given string is quoted inside of a content disposition parameter.
    
    Parameters
    ----------
    string : `str`
        The string to check.

    Returns
    -------
    is_quoted : `bool`
    """
    if (len(string) > 1) and (string[0] == '"') and (string[-1] == '"'):
        return True
    
    return False


def _is_extended_parameter(string):
    """
    Returns whether the given string is an extended parameter of a content disposition parameter.
    
    Parameters
    ----------
    string : `str`
        The string to check.

    Returns
    -------
    is_extended_parameter : `bool`
    """
    return string.endswith('*')


def _is_rfc5987(string):
    """
    Returns whether the given string is an extended notation based on `rfc5987` using `'` signs to capture the language
        parameter.
    
    Parameters
    ----------
    string : `str`
        The string to check.

    Returns
    -------
    is_rfc5987 : `bool`
    """
    return _is_token(string) and string.count("'") == 2


def _unescape(text, *, chars=''.join(map(re.escape, CHARS))):
    """
    Unescapes the given part of a content disposition parameter.
    
    Parameters
    ----------
    text : `str`
        The string to check.
    chars : `str, Optional (Keyword only)
        Characters to match when unescaping.
    
    Returns
    -------
    is_rfc5987 : `bool`
    """
    return re.sub(f'\\\\([{chars}])', '\\1', text)


def parse_content_disposition(header):
    """
    Parsers the content disposition headers.
    
    Parameters
    ----------
    header : `str` header value.
        The headers to pare content disposition headers from.

    Returns
    -------
    disposition_type : `None`, `str`
        The disposition's type if anything found.
    parameters : `dict` of (`str`, `str`) items
        The parsed out parameters.
    """
    if not header:
        return None, {}
    
    disposition_type, *parts = header.split(';')
    if not _is_token(disposition_type):
        return None, {}
    
    parameters = {}
    while parts:
        item = parts.pop(0)
        
        if ('=' not in item):
            return None, parameters
        
        key, value = item.split('=', 1)
        key = key.lower().strip()
        value = value.lstrip()
        
        if key in parameters:
            return None, parameters
        
        if not _is_token(key):
            continue
        
        elif _is_continuous_parameter(key):
            if _is_quoted(value):
                value = _unescape(value[1:-1])
            elif not _is_token(value):
                continue
        
        elif _is_extended_parameter(key):
            if _is_rfc5987(value):
                encoding, _, value = value.split("'", 2)
                if not encoding:
                    encoding = 'utf-8'
            else:
                continue
            
            try:
                value = unquote(value, encoding, 'strict')
            except UnicodeDecodeError:
                continue
        
        else:
            failed = True
            
            if _is_quoted(value):
                value = _unescape(value[1:-1].lstrip('\\/'))
            elif _is_token(value):
                failed = False
            elif parts:
                joined_value = f'{value};{parts[0]}'
                if _is_quoted(joined_value):
                    parts.pop(0)
                    value = _unescape(joined_value[1:-1].lstrip('\\/'))
                    failed = False
            
            if failed:
                return None, {}
        
        parameters[key] = value
    
    return disposition_type.lower(), parameters


def get_content_disposition_filename(parameters, name='filename'):
    """
    Gets the file's name from content disposition parameters.
    
    Parameters
    ----------
    parameters : `dict` of (`str`, `str`) items
        The content disposition parameters parsed from a header value.
    name : `str` = `'filename'`, Optional
        File name to get.

    Returns
    -------
    filename : `None`, `str`
        The file's name if any.
    """
    if not parameters:
        return None
    
    name_suf = f'{name}*'
    
    try:
        return parameters[name_suf]
    except KeyError:
        pass
    
    try:
        return parameters[name]
    except KeyError:
        pass
    
    parts = []
    file_name_parameters = sorted((key, value) for key, value in parameters.items() if key.startswith('filename*'))
    for index, (key, value) in enumerate(file_name_parameters):
        _, tail = key.split('*', 1)
        if tail.endswith('*'):
            tail = tail[:-1]
        if tail == str(index):
            parts.append(value)
        else:
            break
    
    if not parts:
        return None
    
    value = ''.join(parts)
    if "'" not in value:
        return value
    
    encoding, _, value = value.split("'", 2)
    encoding = encoding or 'utf-8'
    return unquote(value, encoding, 'strict')


class MultipartWriter(PayloadBase):
    """
    Multipart body writer.
    
    Attributes
    ----------
    content_type : `None`, `str`
        The payload's content type.
    data : `list` of ``PayloadBase``
        The contained payloads.
    filename : `None`, `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        Payload specific headers.
    size : `None`, `int`
        The payload's size if applicable.
    _boundary : `bytes`
        Boundary to mark the payload's start and end.
    """
    __slots__ = ('_boundary', )
    
    def __init__(self, subtype='mixed', boundary=None):
        """
        Creates a new ``MultipartWriter`` with the given parameters.
        
        Parameters
        ----------
        subtype : `str` = `'mixed'`, Optional
            The subtype of the multipart writer. Defaults to `'mixed'`, but also can `'form-data'` for example.
        boundary : `Nome`, `str` = `None`, Optional
            Boundary to mark the payload's start and end. If not given or given as `None`, then is autogenerated.
        
        Raises
        ------
        UnicodeEncodeError
            `boundary` is given as `str`, but is not `ascii` encodable.
        ValueError
            `boundary` contains invalid character.
        """
        if (boundary is None):
            boundary = uuid.uuid4().hex.encode('ascii')
        else:
            try:
                boundary = boundary.encode('ascii')
            except UnicodeEncodeError as err:
                raise ValueError(
                    f'`boundary` should contain only ASCII chars, got {boundary!r}.'
                ) from err
        
        # Refer to RFCs 7231, 7230, 5234.
        #
        # parameter = token "=" ( token / quoted-string )
        # token = 1*tchar
        # quoted-string = DQUOTE *( qdtext / quoted-pair ) DQUOTE
        # qdtext = HTAB / SP / %x21 / %x23-5B / %x5D-7E / obs-text
        # obs-text = %x80-FF
        # quoted-pair = "\" ( HTAB / SP / VCHAR / obs-text )
        # tchar = "!" / "#" / "$" / "%" / "&" / "'" / "*"
        #                  / "+" / "-" / "." / "^" / "_" / "`" / "|" / "~"
        #                  / DIGIT / ALPHA
        #                  ; any VCHAR, except delimiters
        # VCHAR = %x21-7E
        
        if VALID_TCHAR_RP.match(boundary) is None:
            if INVALID_QDTEXT_CHAR_RP.search(boundary) is not None:
                raise ValueError(
                    f'Boundary value contains invalid characters; got {boundary!r}.'
                )
            
            # escape %x5C and %x22
            quoted_boundary = boundary.replace(b'\\', b'\\\\')
            quoted_boundary = quoted_boundary.replace(b'"', b'\\"')
            quoted_boundary = quoted_boundary.decode('ascii')
            quoted_boundary = f'"{quoted_boundary}"'
        else:
            quoted_boundary = boundary.decode('ascii')
        
        kwargs = {'content_type': f'multipart/{subtype}; boundary={quoted_boundary}'}
        
        PayloadBase.__init__(self, [], kwargs)
        
        self._boundary = boundary
        self.headers[CONTENT_TYPE] = self.content_type
    
    
    @property
    def boundary(self):
        """
        Returns the multipart writer's boundary as string.
        
        Returns
        -------
        boundary : `str`
        """
        return self._boundary.decode('ascii')
    
    
    def append(self, body_part, headers=None):
        """
        Adds a new body part to the ``MultipartWriter``.
        
        Parameters
        ----------
        body_part : ``PayloadBase``, ``BodyPartReader``, `bytes`, `bytearray`, `memoryview`, `BytesIO`, `StringIO`, \
            `TextIOBase`, `BufferedReader`, `BufferedRandom`, `IOBase`, ``AsyncIO``, `async-iterable`
        headers : `None`, ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items = `None`, Optional
            Optional headers for the field.
        
        Returns
        -------
        payload : ``PayloadBase``
            The created payload.
        
        Raises
        ------
        TypeError
            Cannot create payload from the given `body_part`.
        RuntimeError
            - The `payload`'s content has unknown content-encoding.
            - The `payload`'s content has unknown content-transfer-encoding.
        """
        if headers is None:
            headers = IgnoreCaseMultiValueDictionary()
        
        if isinstance(body_part, PayloadBase):
            if (headers is not None):
                body_part.headers.extend(headers)
            
            payload = body_part
        else:
            kwargs = {}
            if (headers is not None):
                kwargs['headers'] = headers
            
            try:
                payload = create_payload(body_part, kwargs)
            except LookupError as err:
                raise TypeError(
                    f'Cannot create payload from: {body_part!r}.'
                ) from err
        
        self.append_payload(payload)
        return payload
    
    
    def append_payload(self, payload):
        """
        Adds a payload to the multipart writer.
        
        Parameters
        ----------
        payload : ``PayloadBase``
            The payload to add to the body.
        
        Raises
        ------
        RuntimeError
            - The `payload`'s content has unknown content-encoding.
            - The `payload`'s content has unknown content-transfer-encoding.
        """
        # content-type
        payload_headers = payload.headers
        if CONTENT_TYPE not in payload_headers:
            payload_headers[CONTENT_TYPE] = payload.content_type
        
        # content-encoding or compression
        try:
            content_encoding = payload_headers[CONTENT_ENCODING].lower()
        except KeyError:
            content_encoding = None
        else:
            if content_encoding in ('deflate', 'gzip', 'br', ):
                pass
            elif content_encoding in ('', 'identity'):
                content_encoding = None
            else:
                raise RuntimeError(
                    f'Unknown content-encoding: {content_encoding!r}.'
                )
        
        # transfer-encoding
        try:
            transfer_encoding = payload_headers[CONTENT_TRANSFER_ENCODING].lower()
        except KeyError:
            transfer_encoding = None
        else:
            if transfer_encoding == '':
                transfer_encoding = None
            elif transfer_encoding in ('base64', 'quoted-printable'):
                pass
            elif transfer_encoding == 'binary':
                transfer_encoding = None
            else:
                raise RuntimeError(
                    f'Unknown content transfer encoding: {transfer_encoding!r}.'
                )
        
        # Set size to payload headers if applicable.
        size = payload.size
        if (size is not None) and (content_encoding is None) and (transfer_encoding is None):
            payload_headers[CONTENT_LENGTH] = str(size)
        
        # Render headers.
        result = []
        extend = result.extend
        for k, v in payload_headers.items():
            extend((k, ': ', v, '\r\n'))
        
        result.append('\r\n')
        
        headers = ''.join(result).encode('utf-8')
        
        # Calculate new size.
        data = self.data
        if data:
            size = self.size
        else:
            size = 6 + len(self._boundary)
            # b'--' + self._boundary + b'--\r\n'
        
        if (size is not None):
            if (content_encoding is not None) or (transfer_encoding is not None) or (payload.size is None):
                size = None
            else:
                size += 6 + len(self._boundary) + payload.size + len(headers)
                # b'--' + self._boundary + b'\r\n' ... data ... b'\r\n'
            
            self.size = size
        
        # Add part.
        data.append((payload, headers, content_encoding, transfer_encoding))
    
    
    def append_json(self, obj, headers=None):
        """
        Helper method to add a json field.
        
        Parameters
        ----------
        obj : `None`, `str`, `int`, `float`, `list` of repeat, `dict` of (`str`, repeat) items
            The payload's data.
        headers : `None`, ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items = `None`, Optional
            Optional headers for the json field.

        Returns
        -------
        payload : ``JsonPayload``
            The created json payload.
        
        Raises
        ------
        RuntimeError
            - The `payload`'s content has unknown content-encoding.
            - The `payload`'s content has unknown content-transfer-encoding.
        """
        kwargs = {}
        if (headers is not None):
            kwargs['headers'] = headers
        
        payload = JsonPayload(obj, kwargs)
        self.append_payload(payload)
        return payload
    
    
    def append_form(self, obj, headers=None):
        """
        Helper method to add url_encoded field.
        
        Parameters
        ----------
        obj : `mapping` of (`str`, `Any`) items, `sequence` of `tuple` (`str`, `Any`) items
            The object, what should be percent encoded for a post request.
        headers : `None`, ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items = `None`, Optional
            Optional headers for the url_encoded field.
        
        Returns
        -------
        payload : ``StringPayload``
            The created string payload.
        
        Raises
        ------
        RuntimeError
            - The `payload`'s content has unknown content-encoding.
            - The `payload`'s content has unknown content-transfer-encoding.
        """
        if hasattr(obj.__class__, 'items'): # mapping type
            obj = list(obj.items())
        
        data = url_encode(obj, doseq=True)
        
        kwargs = {'content_type': 'application/x-www-form-url_encoded'}
        
        if (headers is not None):
            kwargs['headers'] = headers
        
        payload = StringPayload(data, kwargs)
        self.append_payload(payload)
        return payload
    
    
    async def write(self, writer, close_boundary=True):
        """
        Writes the payloads of the multipart writer to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        close_boundary : `bool` = `True`, Optional
            Whether the multipart's payload should be closed with it's boundary.
        """
        for part, headers, content_encoding, transfer_encoding in self.data:
            await writer.write(b'--' + self._boundary + b'\r\n') # fb strings pls!
            await writer.write(headers)
            
            if (content_encoding is not None) or (transfer_encoding is not None):
                multipart_payload_writer = MultipartPayloadWriter(writer, content_encoding, transfer_encoding)
                
                await part.write(multipart_payload_writer)
                await multipart_payload_writer.write_eof()
            else:
                await part.write(writer)
            await writer.write(b'\r\n')
        
        if close_boundary:
            await writer.write(b'--' + self._boundary + b'--\r\n')


TRANSFER_ENCODING_NONE = 0
TRANSFER_ENCODING_BASE64 = 1
TRANSFER_ENCODING_QUOTED_PRINTABLE = 2

class MultipartPayloadWriter:
    """
    Multipart payload writer of ``MultipartWriter``.
    
    Attributes
    ----------
    compressor : `None`, `ZLIB_COMPRESSOR`, `BROTLI_COMPRESSOR`
        The compressor matching the respective payload's content-encoding.
         used by the respective payload
    encoding_buffer : `None`, `bytearray`
        Buffer used when transfer-encoding is base64.
    transfer_encoding : `int`
        The transfer encoding's identifier used by the multipart writer.
        
        Can be any of the following values:
        
        +---------------------------------------+-------+
        | Respective name                       | Value |
        +---------------------------------------+-------+
        | TRANSFER_ENCODING_NONE                | `0`   |
        +---------------------------------------+-------+
        | TRANSFER_ENCODING_BASE64              | `1`   |
        +---------------------------------------+-------+
        | TRANSFER_ENCODING_QUOTED_PRINTABLE    | `2`   |
        +---------------------------------------+-------+
    writer : ``HTTPStreamWriter``
        HTTP writer to write the encoded data into.
    """
    __slots__ = ('compressor', 'encoding_buffer', 'transfer_encoding', 'writer',)
    
    def __new__(cls, writer, content_encoding, transfer_encoding):
        """
        Creates a new ``MultipartPayloadWriter`` with the given parameters.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            HTTP writer to write the encoded data into.
        content_encoding : `None`, `str`
            Content encoding to write the data with.
            
            Can be any of the following:
            
            +---------------+-----------------------------------------------+
            | Value         | Used compressor                               |
            +===============+===============================================+
            | `None`        | `None`                                        |
            +---------------+-----------------------------------------------+
            | `'gzip'`      | `ZLIB_COMPRESSOR(wbits=16 + ZLIB_MAX_WBITS)`  |
            +---------------+-----------------------------------------------+
            | `'deflate'`   | `ZLIB_COMPRESSOR(wbits=-ZLIB_MAX_WBITS)`      |
            +---------------+-----------------------------------------------+
            | `'br'`        | `BROTLI_COMPRESSOR()`                         |
            +---------------+-----------------------------------------------+
            | `'identity'`  | `None`                                        |
            +---------------+-----------------------------------------------+
            
            You need to have `brotlipy` installed to handle `'br'` encoding.
        
        transfer_encoding : `None`, `str`
            Transfer encoding to write the data with.
            
            Can be given as any of:
            +-----------------------+---------------------------------------+
            | Value                 | Used transfer-encoding                |
            +=======================+=======================================+
            | `None`                | TRANSFER_ENCODING_NONE                |
            +-----------------------+---------------------------------------+
            | `'base64'`            | TRANSFER_ENCODING_BASE64              |
            +-----------------------+---------------------------------------+
            | `'quoted-printable'`  | TRANSFER_ENCODING_QUOTED_PRINTABLE    |
            +-----------------------+---------------------------------------+
            | ...                   | TRANSFER_ENCODING_NONE                |
            +-----------------------+---------------------------------------+
        
        Raises
        ------
        ContentEncodingError
            If the given ``content_encoding`` is not supported.
        """
        if content_encoding is None:
            compressor = None
        
        elif content_encoding == 'gzip':
            compressor = ZLIB_COMPRESSOR(wbits=16 + ZLIB_MAX_WBITS)
        
        elif content_encoding == 'deflate':
            compressor = ZLIB_COMPRESSOR(wbits=-ZLIB_MAX_WBITS)
        
        elif content_encoding == 'br':
            if BROTLI_COMPRESSOR is None:
                raise ContentEncodingError(
                    'Can not decode content-encoding: brotli (br). Please install `brotlipy`.'
                )
            
            compressor = BROTLI_COMPRESSOR()
        
        elif content_encoding == 'identity':
            # I assume this is no encoding
            compressor = None
        
        else:
            raise ContentEncodingError(
                f'Can not decode content-encoding: {content_encoding!r}.'
            )
        
        
        if transfer_encoding is None:
            transfer_encoding = TRANSFER_ENCODING_NONE
            encoding_buffer = None
            
        elif transfer_encoding == 'base64':
            transfer_encoding = TRANSFER_ENCODING_BASE64
            encoding_buffer = bytearray()
            
        elif transfer_encoding == 'quoted-printable':
            transfer_encoding = TRANSFER_ENCODING_QUOTED_PRINTABLE
            encoding_buffer = None
        
        else:
            transfer_encoding = TRANSFER_ENCODING_NONE
            encoding_buffer = None
        
        self = object.__new__(cls)
        self.writer = writer
        self.transfer_encoding = transfer_encoding
        self.compressor = compressor
        self.encoding_buffer = encoding_buffer
        return self
    
    
    async def write_eof(self):
        """
        Writes eof to the writer ending the ``MultipartPayloadWriter``, but not to the ``HTTPStreamWriter``.
        
        Writing eof at this case, means, to end the respective payload part, what the multipart payload writer actually
        writes.
        
        This method is a coroutine.
        """
        compressor = self.compressor
        if (compressor is not None):
            self.compressor = None
            chunk = compressor.flush()
            
            if chunk:
                await self.write(chunk)
        
        if self.transfer_encoding == TRANSFER_ENCODING_BASE64:
            encoding_buffer = self.encoding_buffer
            if encoding_buffer:
                await self.writer.write(base64.b64encode(encoding_buffer))
    
    
    async def write(self, chunk):
        """
        Writes a chunk of data to the ``MultipartPayloadWriter``.
        
        This method is a coroutine.
        
        Parameters
        ----------
        chunk : `bytes-like`
            A chunk of data to write.
        """
        compressor = self.compressor
        if (compressor is not None):
            if chunk:
                chunk = compressor.compress(chunk)
                if not chunk:
                    return
        
        transfer_encoding = self.transfer_encoding
        if transfer_encoding == TRANSFER_ENCODING_BASE64:
            encoding_buffer = self.encoding_buffer
            encoding_buffer.extend(chunk)
            
            barrier = len(encoding_buffer)
            if not barrier:
                return
            
            barrier = (barrier // 3) * 3
            if not barrier:
                return
            
            chunk = encoding_buffer[:barrier]
            del encoding_buffer[:barrier]
            chunk = base64.b64encode(chunk)
        
        elif transfer_encoding == TRANSFER_ENCODING_QUOTED_PRINTABLE:
            chunk = binascii.b2a_qp(chunk)
            
        else:
            pass
        
        await self.writer.write(chunk)



from .console import *
from .console_helpers import *
from .editors import *
from .history import *


__all__ = (
    *console.__all__,
    *console_helpers.__all__,
    *editors.__all__,
    *history.__all__,
)



__all__ = ()

from ..utils import IgnoreCaseString


METHOD_ANY = '*'
METHOD_CONNECT = 'CONNECT'
METHOD_HEAD = 'HEAD'
METHOD_GET = 'GET'
METHOD_DELETE = 'DELETE'
METHOD_OPTIONS = 'OPTIONS'
METHOD_PATCH = 'PATCH'
METHOD_POST = 'POST'
METHOD_PUT = 'PUT'
METHOD_TRACE = 'TRACE'

METHOD_GET_ALL = frozenset((METHOD_GET, METHOD_HEAD, METHOD_OPTIONS))
METHOD_POST_ALL = frozenset((METHOD_PATCH, METHOD_POST, METHOD_PUT))

METHOD_ALL = frozenset((METHOD_CONNECT, METHOD_HEAD, METHOD_GET, METHOD_DELETE, METHOD_OPTIONS, METHOD_PATCH,
    METHOD_POST, METHOD_PUT, METHOD_TRACE))


ACCEPT = IgnoreCaseString('Accept')
ACCEPT_CHARSET = IgnoreCaseString('Accept-Charset')
ACCEPT_ENCODING = IgnoreCaseString('Accept-Encoding')
ACCEPT_LANGUAGE = IgnoreCaseString('Accept-Language')
ACCEPT_RANGES = IgnoreCaseString('Accept-Ranges')
ACCESS_CONTROL_MAX_AGE = IgnoreCaseString('Access-Control-Max-Age')
ACCESS_CONTROL_ALLOW_CREDENTIALS = IgnoreCaseString('Access-Control-Allow-Credentials')
ACCESS_CONTROL_ALLOW_HEADERS = IgnoreCaseString('Access-Control-Allow-Headers')
ACCESS_CONTROL_ALLOW_METHODS = IgnoreCaseString('Access-Control-Allow-Methods')
ACCESS_CONTROL_ALLOW_ORIGIN = IgnoreCaseString('Access-Control-Allow-Origin')
ACCESS_CONTROL_EXPOSE_HEADERS = IgnoreCaseString('Access-Control-Expose-Headers')
ACCESS_CONTROL_REQUEST_HEADERS = IgnoreCaseString('Access-Control-Request-Headers')
ACCESS_CONTROL_REQUEST_METHOD = IgnoreCaseString('Access-Control-Request-Method')
AGE = IgnoreCaseString('Age')
ALLOW = IgnoreCaseString('Allow')
AUTHORIZATION = IgnoreCaseString('Authorization')
CACHE_CONTROL = IgnoreCaseString('Cache-Control')
CONNECTION = IgnoreCaseString('Connection')
CONTENT_DISPOSITION = IgnoreCaseString('Content-Disposition')
CONTENT_ENCODING = IgnoreCaseString('Content-Encoding')
CONTENT_LANGUAGE = IgnoreCaseString('Content-Language')
CONTENT_LENGTH = IgnoreCaseString('Content-Length')
CONTENT_LOCATION = IgnoreCaseString('Content-Location')
CONTENT_MD5 = IgnoreCaseString('Content-MD5')
CONTENT_RANGE = IgnoreCaseString('Content-Range')
CONTENT_TRANSFER_ENCODING = IgnoreCaseString('Content-Transfer-Encoding')
CONTENT_TYPE = IgnoreCaseString('Content-Type')
COOKIE = IgnoreCaseString('Cookie')
DATE = IgnoreCaseString('Date')
DESTINATION = IgnoreCaseString('Destination')
DIGEST = IgnoreCaseString('Digest')
ETAG = IgnoreCaseString('Etag')
EXPECT = IgnoreCaseString('Expect')
EXPIRES = IgnoreCaseString('Expires')
FORWARDED = IgnoreCaseString('Forwarded')
FROM = IgnoreCaseString('From')
HOST = IgnoreCaseString('Host')
IF_MATCH = IgnoreCaseString('If-Match')
IF_MODIFIED_SINCE = IgnoreCaseString('If-Modified-Since')
IF_NONE_MATCH = IgnoreCaseString('If-None-Match')
IF_RANGE = IgnoreCaseString('If-Range')
IF_UNMODIFIED_SINCE = IgnoreCaseString('If-Unmodified-Since')
KEEP_ALIVE = IgnoreCaseString('Keep-Alive')
LAST_EVENT_ID = IgnoreCaseString('Last-Event-ID')
LAST_MODIFIED = IgnoreCaseString('Last-Modified')
LINK = IgnoreCaseString('Link')
LOCATION = IgnoreCaseString('Location')
MAX_FORWARDS = IgnoreCaseString('Max-Forwards')
ORIGIN = IgnoreCaseString('Origin')
PRAGMA = IgnoreCaseString('Pragma')
PROXY_AUTHENTICATE = IgnoreCaseString('Proxy-Authenticate')
PROXY_AUTHORIZATION = IgnoreCaseString('Proxy-Authorization')
RANGE = IgnoreCaseString('Range')
REFERER = IgnoreCaseString('Referer')
RETRY_AFTER = IgnoreCaseString('Retry-After')
SEC_WEBSOCKET_ACCEPT = IgnoreCaseString('Sec-WebSocket-Accept')
SEC_WEBSOCKET_VERSION = IgnoreCaseString('Sec-WebSocket-Version')
SEC_WEBSOCKET_PROTOCOL = IgnoreCaseString('Sec-WebSocket-Protocol')
SEC_WEBSOCKET_EXTENSIONS = IgnoreCaseString('Sec-WebSocket-Extensions')
SEC_WEBSOCKET_KEY = IgnoreCaseString('Sec-WebSocket-Key')
SEC_WEBSOCKET_KEY1 = IgnoreCaseString('Sec-WebSocket-Key1')
SERVER = IgnoreCaseString('Server')
SET_COOKIE = IgnoreCaseString('Set-Cookie')
TE = IgnoreCaseString('TE')
TRAILER = IgnoreCaseString('Trailer')
TRANSFER_ENCODING = IgnoreCaseString('Transfer-Encoding')
UPGRADE = IgnoreCaseString('Upgrade')
WEBSOCKET = IgnoreCaseString('WebSocket')
URI = IgnoreCaseString('URI')
USER_AGENT = IgnoreCaseString('User-Agent')
VARY = IgnoreCaseString('Vary')
VIA = IgnoreCaseString('Via')
WANT_DIGEST = IgnoreCaseString('Want-Digest')
WARNING = IgnoreCaseString('Warning')
WWW_AUTHENTICATE = IgnoreCaseString('WWW-Authenticate')
X_FORWARDED_FOR = IgnoreCaseString('X-Forwarded-For')
X_FORWARDED_HOST = IgnoreCaseString('X-Forwarded-Host')
X_FORWARDED_PROTO = IgnoreCaseString('X-Forwarded-Proto')



__all__ = ()

import codecs, re
from http.cookies import CookieError, SimpleCookie

from ..core import Task
from ..utils import from_json
from ..web_common.headers import CONNECTION, CONTENT_TYPE, METHOD_HEAD, SET_COOKIE
from ..web_common.helpers import HttpVersion10
from ..web_common.multipart import MimeType


try:
    import cchardet as chardet
except ImportError:
    try:
        import chardet
    except ImportError as err:
        chardet = None

JSON_RE = re.compile(r'^application/(?:[\w.+-]+?\+)?json')


class ClientResponse:
    """
    Http response class used by ``HTTPClient``.
    
    Attributes
    ----------
    _released : `bool`
        Whether the connection is released.
    body : `None`, `bytes`
        The received response body. Set as `None` if the response body is not yet received, or if it is empty.
    closed : `bool`
        Whether the response is closed.
    connection : `None`, ``Connection``
        Connection used to receive the request response. Set as `None` if the response is ``.close``-d or
        ``.release``-d.
    payload_waiter : `None`, ``Future``
        Future used to retrieve the response's body. It's result is set, when the respective protocol's reader task
        finished.
    cookies : `http.cookies.SimpleCookie`
        Received cookies with the response.
    headers : `None`, ``IgnoreCaseMultiValueDictionary``
        Headers of the response. Set when the http response is successfully received.
    history : `None`, `tuple` of ``ClientResponse``
        Response history. Set as `tuple` of responses from outside.
    loop : ``EventThread``
        The event loop, trough what the request is executed.
    method : `str`
        Method of the respective request.
    status : `None`, `int`
        Received status code. Set as `0` by default.
    url : ``URL``
        The requested url.
    writer : ``Task`` of ``ClientRequest.write_bytes``
        Payload writer task of the respective request.
    raw_message : `None`, ``RawResponseMessage``
        Raw received http response.
    """
    __slots__ = (
        '_released', 'body', 'closed', 'connection', 'payload_waiter', 'cookies', 'headers', 'history', 'loop',
        'method', 'status', 'url', 'writer', 'raw_message'
    )
       
    def __new__(cls, request, connection):
        """
        Crates a new ``ClientResponse`` from the given request and connection.
        
        Parameters
        ----------
        request : ``ClientRequest``
            The respective request.
        connection : ``Connection``
            The connection used to send the request and receive the response.
        """
        self = object.__new__(cls)
        self.loop = request.loop
        self.method = request.method
        self.url = request.original_url
        
        self.writer = request.writer
        self.closed = False
        self.cookies = SimpleCookie()
        self._released = False
        
        self.body = None
        self.status = 0
        self.payload_waiter = None
        self.headers = None
        self.connection = connection
        
        self.raw_message = None
        self.history = None  # will be added later
        
        return self
    
    
    @property
    def reason(self):
        """
        Returns the server response reason.
        
        Returns
        -------
        reason : `None`, `str`
        """
        message = self.raw_message
        if (message is not None):
            reason = message.reason
            if (reason is not None):
                return reason.decode()
    
    
    def __del__(self):
        """releases the response if not yet closed."""
        if self.closed:
            return
        
        self._release_connection()
    
    
    def __repr__(self):
        """Returns the response's representation."""
        ascii_encodable_url = str(self.url)
        
        return f'<{self.__class__.__name__}({ascii_encodable_url}) [{self.status} {self.reason!r}]>'
    
    
    async def start(self,):
        """
        Starts response processing.
        
        This method is a coroutine.
        
        Returns
        -------
        self : ``ClientResponse``
        """
        try:
            protocol = self.connection.protocol
            
            payload_waiter = protocol.set_payload_reader(protocol._read_http_response())
            self.raw_message = message = await payload_waiter
            
            if self.method == METHOD_HEAD:
                payload_reader = None
            else:
                payload_reader = protocol.get_payload_reader_task(message)
            
            if (payload_reader is None):
                payload_waiter = None
                self._response_eof(None)
            else:
                payload_waiter = protocol.set_payload_reader(payload_reader)
                protocol.handle_payload_waiter_cancellation()
                payload_waiter.add_done_callback(self._response_eof)
            
            # response status
            self.status = message.status
            # headers
            self.headers = message.headers
            # OwO
            self.payload_waiter = payload_waiter
            
            # cookies
            for header in self.headers.get_all(SET_COOKIE, ()):
                try:
                    self.cookies.load(header)
                except CookieError: # so sad
                    pass
        except:
            self.close()
            raise
        
        return self
    
    
    def _response_eof(self, future):
        """
        Future callback added to the payload waiter future, to release the used connection.
        
        Parameters
        ----------
        future : ``Future``
            ``.payload_waiter`` future.
        """
        if self.closed:
            return
        
        self.payload_waiter = None
        
        connection = self.connection
        if (connection is not None):
            # WebSocket, protocol could be `None`, because connection could be detached.
            if (connection.protocol is not None) and self.raw_message.upgraded:
                return
            
            self._release_connection()
        
        self.closed = True
        self._cleanup_writer()
    
    
    def _release_connection(self):
        """
        Releases the response's connection.
        
        If the connection type is "close", closes the protocol as well.
        """
        connection = self.connection
        if connection is None:
            return
        
        headers = self.headers
        if (headers is None):
            connection_type = None
        else:
            connection_type = headers.get(CONNECTION, None)
        
        if (
            (self.raw_message.version <= HttpVersion10) if
            (connection_type is None) else
            (connection_type.lower() == 'close')
        ):
            connection.close()
        else:
            connection.release()
        self.connection = None
    
    
    def _notify_content(self):
        """
        Called when response reading is cancelled or released. Sets `ConnectionError` to the respective protocol if
        the payload is still reading.
        """
        payload_waiter = self.payload_waiter
        if (payload_waiter is not None):
            connection = self.connection
            if (connection is not None):
                connection.protocol.set_exception(ConnectionError('Connection closed.'))
        
        self._released = True
    
    
    def _cleanup_writer(self):
        """
        Cancels the writer task of the respective request. Called when the response is cancelled or released, or if
        reading the whole response is done.
        """
        writer = self.writer
        if (writer is not None):
            self.writer = None
            writer.cancel()
    
    
    async def read(self):
        """
        Reads the response's body.
        
        This method is a coroutine.
        
        Returns
        -------
        body : `bytes`
        """
        payload_waiter = self.payload_waiter
        if (payload_waiter is None):
            body = self.body
        else:
            try:
                self.body = body = await payload_waiter
            finally:
                self.payload_waiter = None
        
        return body
    
    
    def get_encoding(self):
        """
        Gets the encoding of the response's body.
        
        Returns
        -------
        encoding : `str`
            Defaults to `'utf-8'`.
        """
        content_type = self.headers.get(CONTENT_TYPE, '').lower()
        mime_type = MimeType(content_type)
        
        encoding = mime_type.parameters.get('charset', None)
        if encoding is not None:
            try:
                codecs.lookup(encoding)
            except LookupError:
                encoding = None
        
        if encoding is None:
            if mime_type.type == 'application' and mime_type.sub_type == 'json':
                encoding = 'utf-8' # RFC 7159 states that the default encoding is UTF-8.
            else:
                if chardet is None:
                    encoding = 'utf-8'
                else:
                    encoding = chardet.detect(self.body)['encoding']
        
        if not encoding:
            encoding = 'utf-8'
        
        return encoding
    
    
    async def text(self, encoding=None, errors='strict'):
        """
        Loads the response's content as text.
        
        This method is a coroutine.
        
        Parameters
        ----------
        encoding : `None`, `str` = `None`, Optional
            If no encoding is given, then detects it from the payload-
        errors : `str` = `'strict'`, Optional
            May be given to set a different error handling scheme. The default `errors` value is `'strict'`, meaning
            that encoding errors raise a `UnicodeError`. Other possible values are `'ignore'`, `'replace'`,
            `'xmlcharrefreplace'`, `'backslashreplace'` and any other name registered via `codecs.register_error()`.
        
        Returns
        -------
        text : `str`
        """
        body = await self.read()
        if body is None:
            return
        
        if encoding is None:
            encoding = self.get_encoding()
        
        return body.decode(encoding, errors)
    
    
    async def json(self, encoding=None, loader=from_json, content_type=None):
        """
        Loads the response's content as a json.
        
        This method is a coroutine.
        
        Parameters
        ----------
        encoding : None`, `str` = `None`, Optional
            Encoding to use instead of the response's. If given as `None` (so by default), then will use the response's
            own encoding.
        loader : `callable` = ``from_json``, Optional
            Json loader. Defaults to json.loads`.
        content_type : `None`, `str` = `None`, Optional
            Content type to use instead of the default one. Defaults to `'application/json'`.
        
        Returns
        -------
        json : `Any`
        
        Raises
        ------
        TypeError
            If the response's mime_type do not match.
        """
        body = await self.read()
        if body is None:
            return
        
        if content_type is not None:
            received_content_type = self.headers.get(CONTENT_TYPE, '').lower()
            
            if (
                (JSON_RE.match(received_content_type) is None)
                     if (content_type is None) else
                (content_type not in received_content_type)
            ):
                raise TypeError(
                    f'Attempt to decode JSON with unexpected mime_type: {received_content_type!r}.'
                )
        
        stripped = body.strip()
        if not stripped:
            return None
        
        if encoding is None:
            encoding = self.get_encoding()
        
        return loader(stripped.decode(encoding))
    
    
    async def __aenter__(self):
        """
        Enters the client response as an asynchronous context manager.
        
        This method is a coroutine.
        """
        return self
    
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Releases the response if not yet closed.
        
        This method is a coroutine.
        """
        self.release()
        return False
    
    
    def close(self):
        """
        Closes the response and it's connection. The used connection will not be reused after.
        """
        if not self._released:
            self._notify_content()
        
        if self.closed:
            return
        
        self.closed = True
        
        connection = self.connection
        if (connection is not None):
            self.connection = None
            connection.close()
        
        self._cleanup_writer()
    
    
    def release(self):
        """
        Releases the response and it's connection. The used connection might be reused after.
        """
        if not self._released:
            self._notify_content()
        
        if self.closed:
            return
        
        self.closed = True
        
        self._release_connection()
        self._cleanup_writer()



__all__ = ()

import re

from .quoting import quote


_TOKEN_RP = re.compile(r'[-!#$%&\'*+.^_`|~0-9a-zA-Z]+')
_LIST_START_RP = re.compile('[\t ,]*')
_SPACE_RP = re.compile('[\t ]*')
_PROTOCOL_RP = re.compile(r'[-!#$%&\'*+.^_`|~0-9a-zA-Z]+(?:/[-!#$%&\'*+.^_`|~0-9a-zA-Z]+)?')

CHARS = frozenset((chr(i) for i in range(0, 128)))
CONTROLS = frozenset((*(chr(i) for i in range(0, 32)), chr(127)))
SEPARATORS = frozenset(('(', ')', '<', '>', '@', ', ', ';', ':', '\\', '"', '/', '[', ']', '?', '=', '{', '}', ' ',
    chr(9)))
TOKENS = CHARS ^ CONTROLS ^ SEPARATORS

def build_extensions(available_extensions):
    """
    Builds websocket extensions header from the given extension values.
    
    Parameters
    ----------
    available_extensions : `list` of `Any`
        Each websocket extension should have the following `4` attributes / methods:
        - `name`: `str`. The extension's name.
        - `request_params` : `list` of `tuple` (`str`, `str`). Additional header parameters of the extension.
        - `decode` : `callable`. Decoder method, what processes a received websocket frame. Should accept `2`
            parameters: The respective websocket ``Frame``, and the ˙max_size` as `int`, what decides the
            maximal size of a received frame. If it is passed, ``PayloadError`` is raised.
        - `encode` : `callable`. Encoder method, what processes the websocket frames to send. Should accept `1`
            parameter, the respective websocket ``Frame``.
    
    Returns
    -------
    header_value : `str`
    """
    main_parts = []
    sub_parts = []
    
    for available_extension in available_extensions:
        name = available_extension.name
        parameters = available_extension.request_params
        
        sub_parts.append(name)
        for key, value in parameters:
            if value is None:
                sub_parts.append(key)
            else:
                sub_parts.append(f'{key}={value}')
        
        main_parts.append('; '.join(sub_parts))
        sub_parts.clear()
    
    return ', '.join(main_parts)


def parse_extensions(header_value):
    """
    Parses extension header.
    
    Parameters
    ----------
    header_value : `str`
        Received extension header.

    Returns
    -------
    result : `list` of `tuple` (`str`, `list` of `tuple` (`str`, `str`))
        The parsed out extensions as `name` - `parameters` pairs. The `parameters` are in `list` storing
        `key` - `value` pairs.
    
    Raises
    ------
    ValueError
        Extension header value is incorrect.
    """
    result = []
    limit = len(header_value)
    index = 0
    
    check_start = True
    
    while True:
        # parse till 1st element
        matched = _LIST_START_RP.match(header_value, index)
        index = matched.end()
        
        # are we at the end?
        if index == limit:
            return result
        
        # now lets parse the extension's name
        matched = _TOKEN_RP.match(header_value, index)
        if matched is None:
            raise ValueError(
                f'Expected extension name since index {index!r}, got {header_value!r}.'
            )
        
        name = matched.group(0)
        index = matched.end()
        
        # nice, we have a name, we can make our item now!
        sub_parts = []
        result.append((name, sub_parts,),)
        
        # should we parse a sublist?
        while True:
            
            # after half item we skip this part
            if check_start:
                # are we at the end?
                if index == limit:
                    return result
                
                # lets parse till next character
                matched = _SPACE_RP.match(header_value, index)
                index = matched.end()
                
                # are we at the end?
                if index == limit:
                    return result
                
                # no sublist?
                if header_value[index] == ',':
                    index += 1
                    break

                # invalid character
                if header_value[index] != ';':
                    raise ValueError(
                        f'Expected \';\' at index {index!r}, got {header_value!r}.'
                    )
                
                # we have a sublist
                index += 1
                
            else:
                check_start = True
            
            # parse space
            matched = _SPACE_RP.match(header_value, index)
            index = matched.end()
            
            # are we at the end?
            if index == limit:
                break
            
            # lets parse the key now
            matched = _TOKEN_RP.match(header_value, index)
            if matched is None:
                raise ValueError(
                    f'Expected parameter name since index {index!r}, got {header_value!r}.'
                )
            
            key = matched.group(0)
            index = matched.end()
            
            # are we at the end?
            if index == limit:
                sub_parts.append((key, None,),)
                break
            
            # parse space
            matched = _SPACE_RP.match(header_value, index)
            index = matched.end()
            
            # are we at the end?
            if index == limit:
                sub_parts.append((key, None,),)
                break

            #is it a full item or a half?
            
            #next extension
            if header_value[index] == ',':
                sub_parts.append((key, None,),)
                index += 1
                break

            # next item
            if header_value[index] == ';':
                sub_parts.append((key, None,),)
                index += 1
                check_start = False
                continue

            #invalid character
            if header_value[index] != '=':
                raise ValueError(
                    f'Expected \',\' or \';\' or \'=\' at index {index!r}, got {header_value!r}.'
                )
            
            index += 1
            
            # parse space
            matched = _SPACE_RP.match(header_value, index)
            index = matched.end()
            
            # are we at the end?
            if index == limit:
                raise ValueError(
                    f'Expected a parameter value, but string ended, got {header_value!r}.'
                )
            
            # is it '"stuff"' ?
            if header_value[index] == '"':
                index += 1
                
                # are we at the end?
                if index == limit:
                    raise ValueError(
                        f'Expected a parameter value, but string ended, got {header_value!r}.'
                    )
                
                matched = _TOKEN_RP.match(header_value, index)
                if matched is None:
                    raise ValueError(
                        f'Expected parameter value since index {index!r}, got {header_value!r}'
                    )
                
                value = matched.group(0)
                index = matched.end()
                
                # are we at the end? or did we finish the string normally?
                if index == limit or header_value[index] != '"':
                    raise ValueError(
                        f'Expected a \'"\' after starting a value with \'"\', got {header_value!r}.'
                    )
                index += 1
            
            # is it 'stuff' ?
            else:
                matched = _TOKEN_RP.match(header_value, index)
                if matched is None:
                    raise ValueError(
                        f'Expected parameter value since index {index!r}, got {header_value!r}.'
                    )
                value = matched.group(0)
                index = matched.end()
            
            # we got a full item
            sub_parts.append((key, value,),)


def parse_connections(header_value):
    """
    Parses subprotocol or connection headers.
    
    Parameters
    ----------
    header_value : `str`
        Received subprotocol or connection header.
    
    Returns
    -------
    result : `list` of `str`
        The parsed subprotocol or connection headers.
    
    Raises
    ------
    ValueError
        Subprotocol or connection header value is incorrect.
    """
    result = []
    limit = len(header_value)
    index = 0
    
    while True:
        #parse till 1st element
        matched = _LIST_START_RP.match(header_value, index)
        index = matched.end()
        
        #are we at the end?
        if index == limit:
            return result
        
        #now lets parse the upgrade's name
        matched = _TOKEN_RP.match(header_value, index)
        if matched is None:
            raise ValueError(
                f'Expected upgrade type since index {index!r}, got {header_value!r}.'
            )
        
        name = matched.group(0)
        index = matched.end()

        #nice
        result.append(name)
        
        #are we at the end?
        if index == limit:
            return result

        #lets parse till next character
        matched = _SPACE_RP.match(header_value, index)
        index = matched.end()

        #are we at the end?
        if index == limit:
            return result
    
        #no sublist?
        if header_value[index] == ',':
            index += 1
            continue
        
        raise ValueError(
            f'Expected \',\' at index {index!r}, got {header_value!r}.'
        )


def build_subprotocols(subprotocols):
    """
    Builds websocket subprotocol headers from the given subprotocol values.
    
    Parameters
    ----------
    subprotocols : `list` of `str`
        A list of supported subprotocols.
    
    Returns
    -------
    header_value : `str`
    """
    return ', '.join(subprotocols)


parse_subprotocols = parse_connections # yes, these are the same


def parse_upgrades(header_value):
    """
    Parses upgrade headers.
    
    Parameters
    ----------
    header_value : `str`
        Received upgrade header.
    
    Returns
    -------
    result : `list` of `str`
        The parsed upgrade headers.
    
    Raises
    ------
    ValueError
        Upgrade header value is incorrect.
    """
    result = []
    limit = len(header_value)
    index = 0
    
    while True:
        # parse till 1st element
        matched = _LIST_START_RP.match(header_value, index)
        index = matched.end()
        
        # are we at the end?
        if index == limit:
            return result
        
        # now lets parse the upgrade's name
        matched = _PROTOCOL_RP.match(header_value, index)
        if matched is None:
            raise ValueError(
                f'Expected upgrade type since index {index!r}, got {header_value!r}.'
            )
        name = matched.group(0)
        index = matched.end()
        
        # nice
        result.append(name)
        
        # are we at the end?
        if index == limit:
            return result
        
        # lets parse till next character
        matched = _SPACE_RP.match(header_value, index)
        index = matched.end()
        
        # are we at the end?
        if index == limit:
            return result
        
        # no sublist?
        if header_value[index] == ',':
            index += 1
            continue
        
        raise ValueError(
            f'Expected \',\' at index {index!r}, got {header_value!r}.'
        )


def build_content_disposition_header(disposition_type, parameters, quote_fields):
    """
    Creates Content-Disposition header value.
    
    Parameters
    ----------
    disposition_type : `str`
        Disposition type. Can be one of following: `'inline'`, `'attachment'`, '`form-data`'.
    parameters : `dict` of (`str`, `str`) items
        Disposition parameters.
    quote_fields : `bool`
        Whether field values should be quoted.
    
    Returns
    -------
    value : `str`
    """
    if (not disposition_type) or not (TOKENS > set(disposition_type)):
        raise ValueError(
            f'Bad content disposition type {disposition_type!r}.'
        )
    
    if parameters:
        parameter_parts = [disposition_type]
        for key, value in parameters.items():
            if (not key) or (not (TOKENS > set(key))):
                raise ValueError(
                    f'Bad content disposition parameter {key!r}={value!r}.'
                )
            
            if quote_fields:
                value = quote(value, '[]')
            
            parameter_parts.append('; ')
            
            parameter_parts.append(key)
            parameter_parts.append('="')
            parameter_parts.append(value)
            parameter_parts.append('"')
            
            if key == 'filename':
                parameter_parts.append('; filename*=utf-8\'\'')
                parameter_parts.append(value)
        
        value = ''.join(parameter_parts)
    else:
        value = disposition_type
    
    return value



__all__ = ('WaitTillAll',)

from .wait_till_first import WaitTillFirst


class WaitTillAll(WaitTillFirst):
    """
    A future subclass, which waits till all the given tasks or futures become done. When finished, returns the `done`
    and the `pending` futures.
    
    Attributes
    ----------
    _blocking : `bool`
        Whether the future is already being awaited, so it blocks the respective coroutine.
    _callbacks : `list` of `callable`
        The callbacks of the future, which are queued up on the respective event loop to be called, when the future is
        finished. These callback should accept `1` parameter, the future itself.
        
        Note, if the future is already done, then the newly added callbacks are queued up instantly on the respective
        event loop to be called.
    
    _exception : `None`, `BaseException`
        The exception set to the future as it's result. Defaults to `None`.
    _loop : ``EventThread``
        The loop to what the created future is bound.
    _result : `tuple` (`set` of ``Future``, `set` of ``Future``)
        The result of the future. Defaults to `None`.
    _state : `str`
        The state of the future.
        
        Can be set as one of the following:
        
        +---------------------------+-----------+
        | Respective name           | Value     |
        +===========================+===========+
        | FUTURE_STATE_PENDING      | `0`       |
        +---------------------------+-----------+
        | FUTURE_STATE_CANCELLED    | `1`       |
        +---------------------------+-----------+
        | FUTURE_STATE_FINISHED     | `2`       |
        +---------------------------+-----------+
        | FUTURE_STATE_RETRIEVED    | `3`       |
        +---------------------------+-----------+
        
        Note, that states are checked by memory address and not by equality. Also ``FUTURE_STATE_RETRIEVED`` is used only if
        `__debug__` is set as `True`.
    
    _callback : ``._wait_callback``
        Callback added to the waited futures.
    """
    __slots__ = ()
    # `__new__` is same as ``WaitTillFirst.__new__``
    # `__repr__` is same as ``Future.__repr__``

    class _wait_callback:
        """
        ``WaitTillAll``'s callback put on the future's waited by it.
        
        Attributes
        ----------
        _parent : ``WaitTillAll``
            The parent future.
        """
        __slots__ = ('_parent',)
        
        def __init__(self, parent):
            """
            Creates a new ``WaitTillAll`` callback object with the given parent ``WaitTillAll``.
            
            Parameters
            ----------
            parent : ``WaitTillAll``
                The parent future.
            """
            self._parent = parent
            
        def __call__(self, future):
            """
            The callback, which runs when a waited future is done.
            
            Removes the done future from the parent's `pending` futures and puts it on the `done` ones. If there is no
            more future to wait for, marks the parent ``WaitTillAll`` as finished as well.
            
            Parameters
            ----------
            future : ``Future``
                A done waited future.
            """
            parent = self._parent
            if parent is None:
                return
            
            done, pending = parent._result
            
            pending.remove(future)
            done.add(future)
            
            if pending:
                return
            
            parent._mark_as_finished()
    
    # `futures_done` is same as ``WaitTillFirst.futures_done``
    # `futures_pending` is same as ``WaitTillFirst.futures_pending``
    # `cancel` is same as ``WaitTillFirst.cancel``
    # `is_cancelled` is same as ``Future.is_cancelled``
    # `is_done` is same as ``Future.is_done``
    # `is_pending` is same as ``Future.is_pending``
    # `result` is same as ``Future.result``
    # `exception` is same as ``Future.exception``
    # `add_done_callback` is same as ``Future.add_done_callback``
    # `remove_done_callback` is same as ``Future.remove_done_callback``
    # `set_result` is same as ``WaitTillFirst.set_result``
    # `set_result_if_pending` is same as ``WaitTillFirst.set_result_if_pending``
    # `set_exception` is same as ``WaitTillFirst.set_exception``
    # `set_exception_if_pending` is same as ``WaitTillFirst.set_exception_if_pending``
    # `__iter__` is same as ``Future.__iter__``
    # `__await__` is same as ``Future.__await__``
    # if __debug__:
    #    `__del__` is same as ``Future.__del__``
    #    `__silence__` is same as ``Future.__silence__``
    #    `__silence_callback__` is same as ``Future.__silence_callback__``
    # `cancel_handles` is same as ``Future.cancel_handles``
    # `clear` is same as ``WaitTillFirst.clear``
    # `sleep` is same as ``WaitTillFirst.clear``
    # `sync_wrap` is same as ``Future.cancel_handles``
    # `async_wrap` is same as ``Future.cancel_handles``



from .async_executing import *
from .event import *
from .future import *
from .future_chaining import *
from .future_async_wrapper import *
from .future_sync_wrapper import *
from .gatherer import *
from .handle_cancellers import *
from .locks import *
from .queues import *
from .result_gathering_future import *
from .task import *
from .task_suppression import *
from .task_thread_switcher import *
from .timeouting import *
from .wait_continously import *
from .wait_till_all import *
from .wait_till_exception import *
from .wait_till_first import *

__all__ = (
    *async_executing.__all__,
    *event.__all__,
    *future.__all__,
    *future_chaining.__all__,
    *future_async_wrapper.__all__,
    *future_sync_wrapper.__all__,
    *gatherer.__all__,
    *handle_cancellers.__all__,
    *locks.__all__,
    *queues.__all__,
    *result_gathering_future.__all__,
    *task.__all__,
    *task_suppression.__all__,
    *task_thread_switcher.__all__,
    *timeouting.__all__,
    *wait_continously.__all__,
    *wait_till_all.__all__,
    *wait_till_exception.__all__,
    *wait_till_first.__all__,
)



__all__ = ('AsyncProcess',)

import sys
from socket import socketpair as create_socket_pair
from subprocess import PIPE, Popen, TimeoutExpired

from ..protocols_and_transports import ReadProtocolBase
from ..traps import Future, Task, WaitTillAll, future_or_timeout, skip_poll_cycle

from .subprocess_protocols import SubprocessReadPipeProtocol, SubprocessWritePipeProtocol
from .subprocess_writer import SubprocessWriter


IS_AIX = sys.platform.startswith('aix')
LIMIT = 1 << 16
MAX_READ_SIZE = 262144

PROCESS_EXIT_DELAY_LIMIT = 10


class AsyncProcess:
    """
    Asynchronous process implementation.
    
    Attributes
    ----------
    _alive_file_descriptors : `list` if `int`
        A list of alive file descriptor's respective internal identifier.
        
        Can have the following elements:
        
        +-------------------+-------+
        | Respective name   | Value |
        +===================+=======+
        | stdin             | `0`   |
        +-------------------+-------+
        | stdout            | `1`   |
        +-------------------+-------+
        | stderr            | `2`   |
        +-------------------+-------+
    
    _connection_lost : `bool`
        Whether all the pipes of the ``AsyncProcess`` lost connection.
    _drain_waiter : `None`, ``Future``
        A future, what is used to block the writing task, till it's writen data is drained.
    _exit_waiters : `None`, `set of ``Future``
        Waiter futures which wait for the subprocess to shutdown.
    _extra : `dict` of (`str`, `Any`) items
        Optional transport information.
    _loop : ``EventThread``
        the respective event loop of the async subprocess to what is bound to.
    _paused : `bool`
        Whether the subprocess is paused writing because it hit's the high water mark.
    _pending_calls : `None`, `list` of (`callable`, `tuple` of `Any`)
        Meanwhile the subprocess connection is established, this attribute is set as a list to put connection lost
        and related calls it to with their parameters.
    _subprocess_stderr_protocol : `None`, ``SubprocessReadPipeProtocol``
        Protocol of the stderr pipe if applicable.
    _subprocess_stdin_protocol : `None`, ``SubprocessWritePipeProtocol``
        Protocol of the stdin pipe if applicable.
    _subprocess_stdout_protocol : `None`, ``SubprocessReadPipeProtocol``
        Protocol of the stdout pipe if applicable.
    _closed : `bool`
        Whether subprocess is closed.
    process : `subprocess.Process`
        The internal blocking subprocess object.
    process_id : `int`
        The subprocess identifier.
    return_code : `None`, `int`
        The returned exit code of the subprocess. Set as `None` if not yet applicable.
    stderr : ``ReadProtocolBase``
        Asynchronous stderr implementation.
    stdin : ``SubprocessWriter``
        Asynchronous stdin implementation.
    stdout : ``ReadProtocolBase``
        Asynchronous stdout implementation.
    """
    __slots__ = (
        '_alive_file_descriptors', '_closed', '_connection_lost', '_drain_waiter', '_exit_waiters', '_extra', '_loop',
        '_paused', '_pending_calls', '_subprocess_stderr_protocol', '_subprocess_stdin_protocol',
        '_subprocess_stdout_protocol', 'process', 'process_id', 'return_code', 'stderr', 'stdin', 'stdout'
    )
    
    async def __new__(cls, loop, process_parameters, shell, stdin, stdout, stderr, buffer_size, extra,
            process_open_kwargs):
        """
        Creates a new ``AsyncProcess``.
        
        This method is a coroutine.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the async process is bound to.
        process_parameters : `tuple` of `Any`
            Process parameters to open the subprocess with.
        shell : `bool
            Whether the specified command will be executed through the shell.
        stdin : `file-like`, `subprocess.PIPE`.
            Standard input for the created shell. Defaults to `subprocess.PIPE`.
        stdout : `file-like`, `subprocess.PIPE`, `subprocess.DEVNULL`
            Standard output for the created shell.
        stderr : `file-like`, `subprocess.PIPE`, `subprocess.DEVNULL`, `subprocess.STDOUT`
            Standard error for the created shell.
        buffer_size : `int`
            Will be supplied as the corresponding parameter to the open() function when creating the
            stdin/stdout/stderr pipe file objects:
            
            Expected values:
            
            +---------------+-----------+-----------------------------------------------------------------------+
            | Name          | Value     | Description                                                           |
            +===============+===========+=======================================================================+
            | unbuffered    | `0`       | Read and write are one system call and can return short.              |
            +---------------+-----------+-----------------------------------------------------------------------+
            | line buffered | `1`       | Only usable if `universal_newlines=True`, for example in text mode.   |
            +---------------+-----------+-----------------------------------------------------------------------+
            | buffer size   | `>1`      | Use a buffer of approximately to that value.                          |
            +---------------+-----------+-----------------------------------------------------------------------+
            | default       | `<0`      | use the system default: `io.DEFAULT_BUFFER_SIZE`.                     |
            +---------------+-----------+-----------------------------------------------------------------------+
            
        extra : `None`, `dict` of (`str`, `Any`) items
            Optional transport information.
        process_open_kwargs : `dict` of (`str`, `Any`) items
            Additional parameters to open the process with.
        
        Raises
        ------
        TypeError
            If `process_open_kwargs` contains unexpected key.
        """
        if stdin == PIPE:
            # Use a socket pair for stdin, since not all platforms support selecting read events on the write end of a
            # socket (which we use in order to detect closing of the other end).  Notably this is needed on AIX, and
            # works just fine on other platforms.
            stdin_r, stdin_w = create_socket_pair()
        else:
            stdin_r = stdin
            stdin_w = None
        
        process = None
        
        try:
            process = Popen(process_parameters, shell=shell, stdin=stdin_r, stdout=stdout, stderr=stderr,
                universal_newlines=False, bufsize=buffer_size, **process_open_kwargs)
            
            if (stdin_w is not None):
                stdin_r.close()
                process.stdin = open(stdin_w.detach(), 'wb', buffering=buffer_size)
                stdin_w = None
        except:
            if (process is not None) and (process.poll() is None):
                try:
                    process.kill()
                except ProcessLookupError:
                    pass
            
            raise
        finally:
            if (stdin_w is not None):
                stdin_r.close()
                stdin_w.close()
        
        if extra is None:
            extra = {}
        
        extra['subprocess'] = process
        
        self = object.__new__(cls)
        self._extra = extra
        self._closed = False
        self._loop = loop
        self.process = process
        self.process_id = process.pid
        self.return_code = None
        self._exit_waiters = None
        self._pending_calls = []
        self._subprocess_stdin_protocol = None
        self._subprocess_stdout_protocol = None
        self._subprocess_stderr_protocol = None
        self.stdin = None
        self.stdout = None
        self.stderr = None
        self._paused = False
        self._drain_waiter = None
        self._connection_lost = False
        self._alive_file_descriptors = []
        
        try:
            stdin = process.stdin
            if (stdin is not None):
                subprocess_stdin_protocol = SubprocessWritePipeProtocol(self, 0)
                await loop.connect_write_pipe(subprocess_stdin_protocol, stdin)
                self._subprocess_stdin_protocol = subprocess_stdin_protocol
                stdin_transport = subprocess_stdin_protocol.get_transport()
                if (stdin_transport is not None):
                    self.stdin = SubprocessWriter(loop, stdin_transport, self)
            
            stdout = process.stdout
            if (stdout is not None):
                subprocess_stdout_protocol = SubprocessReadPipeProtocol(self, 1)
                await loop.connect_read_pipe(subprocess_stdout_protocol, stdout)
                self._subprocess_stdout_protocol = subprocess_stdout_protocol
                stdout_transport = subprocess_stdout_protocol.get_transport()
                if (stdout_transport is not None):
                    self.stdout = stdout_protocol = ReadProtocolBase(loop)
                    stdout_protocol.connection_made(stdout_transport)
                    self._alive_file_descriptors.append(1)
            
            stderr = process.stderr
            if (stderr is not None):
                subprocess_stderr_protocol = SubprocessReadPipeProtocol(self, 2)
                await loop.connect_read_pipe(subprocess_stderr_protocol, stderr)
                self._subprocess_stderr_protocol = subprocess_stderr_protocol
                stderr_transport = subprocess_stderr_protocol.get_transport()
                if (stderr_transport is not None):
                    self.stderr = ReadProtocolBase(loop)
                    self.stderr.connection_made(stderr_transport)
                    self._alive_file_descriptors.append(2)
            
            for pending_call, args in self._pending_calls:
                pending_call(self, *args)
            
            self._pending_calls = None
        except:
            self.close()
            await self.wait()
            raise
        
        return self
    
    def __repr__(self):
        """Returns the async process's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        if self._closed:
            repr_parts.append(' closed')
            field_added = True
        else:
            field_added = False
        
        stdin = self.stdin
        if (stdin is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' stdin=')
            repr_parts.append(repr(stdin))
        
        stdout = self.stdout
        if (stdout is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' stdout=')
            repr_parts.append(repr(stdout))
        
        stderr = self.stderr
        if (stderr is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' stderr=')
            repr_parts.append(repr(stderr))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any` = `None`, Optional
            Default value to return if `name` could not be matched. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        return self._extra.get(name, default)
    
    
    def is_closing(self):
        """
        Returns whether the async process is closing.
        
        Returns
        -------
        is_closing : `bool`
        """
        return self._closed
    
    
    def close(self):
        """
        Starts the shutdown process of the async process.
        """
        if self._closed:
            return
        
        self._closed = True
        
        pipe_stdin = self._subprocess_stdin_protocol
        if (pipe_stdin is not None):
            pipe_stdin.close_transport()
        
        pipe_stdout = self._subprocess_stdout_protocol
        if (pipe_stdout is not None):
            pipe_stdout.close_transport()
        
        pipe_stderr = self._subprocess_stderr_protocol
        if (pipe_stderr is not None):
            pipe_stderr.close_transport()
        
        process = self.process
        if (process is not None):
            if (self.return_code is None) and (process.poll() is None):
                try:
                    process.kill()
                except ProcessLookupError:
                    pass
        
        Task(self._process_exited(), self._loop)
    
    __del__ = close
    
    def send_signal(self, signal):
        """
        Sends the signal to the child process.
        
        Parameters
        ----------
        signal : `int`
            The signal to send.
        
        Raises
        ------
        ProcessLookupError
            The underlying process is already dead.
        """
        process = self.process
        if process is None:
            raise ProcessLookupError()
        
        process.send_signal(signal)
    
    
    def terminate(self):
        """
        Stops the child process.
        
        Raises
        ------
        ProcessLookupError
            The underlying process is already dead.
        """
        process = self.process
        if process is None:
            raise ProcessLookupError()
        
        process.terminate()
    
    
    async def kill(self):
        """
        Kills the child process.
        
        This method is a coroutine.
        """
        process = self.process
        if process is None:
            return
        
        process.kill()
        
        self.close()
        
        await skip_poll_cycle(self._loop)
    
    
    def _pipe_connection_lost(self, file_descriptor, exception):
        """
        Called when a file descriptor of the subprocess lost connection.
        
        Calls ``._do_pipe_connection_lost`` or adds it as a callback if the process is still connecting.
        
        Parameters
        ----------
        file_descriptor : `int`
            File descriptor identifier.
            
            It's value can be any of the following:
            
            +-------------------+-------+
            | Respective name   | Value |
            +===================+=======+
            | stdin             | `0`   |
            +-------------------+-------+
            | stdout            | `1`   |
            +-------------------+-------+
            | stderr            | `2`   |
            +-------------------+-------+
        
        exception : `None`, `BaseException`
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        pending_calls = self._pending_calls
        if (pending_calls is None):
            self._do_pipe_connection_lost(file_descriptor, exception)
        else:
            pending_calls.append((self.__class__._do_pipe_connection_lost, (file_descriptor, exception)))
        
        self._try_finish()
    
    
    def _do_pipe_connection_lost(self, file_descriptor, exception):
        """
        Called by ``._pipe_connection_lost`` to call the pipe's connection lost method.
        
        Parameters
        ----------
        file_descriptor : `int`
            File descriptor identifier.
            
            It's value can be any of the following:
            
            +-------------------+-------+
            | Respective name   | Value |
            +===================+=======+
            | stdin             | `0`   |
            +-------------------+-------+
            | stdout            | `1`   |
            +-------------------+-------+
            | stderr            | `2`   |
            +-------------------+-------+
        
        exception : `None`, `BaseException`
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        if file_descriptor == 0:
            pipe = self.stdin
            if (pipe is not None):
                pipe.close()
            
            self._do_connection_lost(exception)
            return
        
        if file_descriptor == 1:
            reader = self.stdout
        elif file_descriptor == 2:
            reader = self.stderr
        else:
            reader = None
        
        if (reader is not None):
            reader.connection_lost(exception)
        
        try:
            self._alive_file_descriptors.remove(file_descriptor)
        except ValueError:
            pass
        
        self._maybe_process_exited()
    
    
    def _pipe_data_received(self, file_descriptor, data):
        """
        Called when one of the subprocess's pipe receives any data.
        
        Calls ``._do_pipe_data_received`` or adds it as a callback if the process is still connecting.
        
        Parameters
        ----------
        file_descriptor : `int`
            File descriptor identifier.
            
            It's value can be any of the following:
            
            +-------------------+-------+
            | Respective name   | Value |
            +===================+=======+
            | stdin             | `0`   |
            +-------------------+-------+
            | stdout            | `1`   |
            +-------------------+-------+
            | stderr            | `2`   |
            +-------------------+-------+
        
        data : `bytes`
            The received data.
        """
        pending_calls = self._pending_calls
        if (pending_calls is None):
            self._do_pipe_data_received(file_descriptor, data)
        else:
            pending_calls.append((self.__class__._do_pipe_data_received, (file_descriptor, data)))
    
    
    def _do_pipe_data_received(self, file_descriptor, data):
        """
        Called by ``._pipe_data_received`` to call the respective protocol' data received method.
        
        Parameters
        ----------
        file_descriptor : `int`
            File descriptor identifier.
            
            It's value can be any of the following:
            
            +-------------------+-------+
            | Respective name   | Value |
            +===================+=======+
            | stdin             | `0`   |
            +-------------------+-------+
            | stdout            | `1`   |
            +-------------------+-------+
            | stderr            | `2`   |
            +-------------------+-------+
        
        data : `bytes`
            The received data.
        """
        if file_descriptor == 1:
            reader = self.stdout
        elif file_descriptor == 2:
            reader = self.stderr
        else:
            return
        
        if (reader is not None):
            reader.data_received(data)
    
    
    async def _process_exited(self):
        """
        Task created by ``.close``to wait till the sub-process is closed and to set
        
        This method is a coroutine.
        """
        try:
            return_code = self.process.poll()
            if return_code is None:
                return_code = await self._loop.run_in_executor(self.process.wait)
            
            self.return_code = return_code
            
            self._try_finish()
            
            # wake up futures waiting for wait()
            exit_waiters = self._exit_waiters
            if (exit_waiters is not None):
                for waiter in self._exit_waiters:
                    if not waiter.is_cancelled():
                        waiter.set_result(return_code)
            
            self._exit_waiters = None
        finally:
            self.process = None
    
    
    def _maybe_process_exited(self):
        """
        Called when a sub-process pipe is closed. When all all pipe is closed, calls ``.close``.
        """
        if self._alive_file_descriptors:
            return
        
        self.close()
    
    
    async def wait(self, timeout=None):
        """
        Wait for child process to terminate.
        
        This method is a coroutine.
        
        Parameters
        ----------
        timeout : `None`, `float`
            The maximal amount of time to wait for the process to close in seconds.

        Returns
        -------
        return_code : `int`
            The return code of the subprocess.
        
        Raises
        ------
        TimeoutExpired
            If the process was not closed before timeout.
        """
        return_code = self.return_code
        if (return_code is not None):
            return return_code
        
        waiter = Future(self._loop)
        
        exit_waiters = self._exit_waiters
        if exit_waiters is None:
            self._exit_waiters = exit_waiters = set()
        
        exit_waiters.add(waiter)
        if (timeout is not None):
            future_or_timeout(waiter, timeout)
        
        try:
            return await waiter
        except TimeoutError:
            exit_waiters.discard(waiter)
            
            process = self.process
            if process is None:
                args = None
            else:
                args = process.args
            
            raise TimeoutExpired(args, timeout) from None
    
    
    def _try_finish(self):
        """
        If the sub-process finished closing, calls ``._do_connection_lost``. If the process is still connecting, adds
        it as a callback instead.
        """
        if self.return_code is None:
            return
        
        stdin_protocol = self._subprocess_stdin_protocol
        if (stdin_protocol is None) or stdin_protocol._disconnected:
            return
        
        stdout_protocol = self._subprocess_stdout_protocol
        if (stdout_protocol is None) or stdout_protocol._disconnected:
            return
        
        stderr_protocol = self._subprocess_stderr_protocol
        if (stderr_protocol is None) or stderr_protocol._disconnected:
            return
        
        pending_calls = self._pending_calls
        if (pending_calls is None):
            self._do_connection_lost(None)
        else:
            pending_calls.append((self.__class__._do_connection_lost, (self, None,)))
    
    
    async def _feed_stdin(self, input_):
        """
        Feeds the given data to ``.stdin``, waits till it drains and closes it.
        
        Used by ``.communicate``.
        
        This method is a coroutine.
        
        Parameters
        ----------
        input_ : `None`, `bytes-like`
            Optional data to be sent to the sub-process.
        """
        stdin = self.stdin
        if stdin is None:
            return
        
        if (input_ is not None):
            stdin.write(input_)
        
        try:
            await stdin.drain()
        except (BrokenPipeError, ConnectionResetError):
            # communicate() ignores BrokenPipeError and ConnectionResetError
            pass
        
        stdin.close()
    
    
    async def _read_close_stdout_stream(self):
        """
        Reads every data from ``.stdout``. When reading is done, closes it.
        
        Used by ``.communicate``.
        
        This method is a coroutine.
        
        Returns
        -------
        result :`None`, `bytes`
        """
        stream = self.stdout
        if stream is None:
            return None
        
        transport = self._subprocess_stdout_protocol.get_transport()
        result = await stream.read()
        transport.close()
        return result
    
    
    async def _read_close_stderr_stream(self):
        """
        Reads every data from ``.stderr``. When reading is done, closes it.
        
        Used by ``.communicate``.
        
        This method is a coroutine.
        
        Returns
        -------
        result : `None`, `bytes`
        """
        stream = self.stderr
        if stream is None:
            return None
        
        transport = self._subprocess_stderr_protocol.get_transport()
        result = await stream.read()
        transport.close()
        return result
    
    
    async def communicate(self, input_=None, timeout=None):
        """
        Sends data to stdin and reads data from stdout and stderr.
        
        Returns when the process is closed or raises when timeout occurs.
        
        This method is a coroutine.
        
        Parameters
        ----------
        input_ : `None`, `bytes-like` = `None` , Optional
            Optional data to be sent to the sub-process.
        timeout : `None`, `float` = `None`, Optional
            The maximal amount of time to wait for the process to close in seconds.
        
        Returns
        -------
        stdout : `None`, `bytes`
            The read data from stdout.
        stderr : `None`, `bytes`
            The read data from stderr.
        
        Raises
        ------
        TimeoutError
            If the process was not closed before timeout.
        """
        tasks = []
        
        loop = self._loop
        if input_ is None:
            stdin_task = None
        else:
            stdin_task = Task(self._feed_stdin(input_), loop)
            tasks.append(stdin_task)
        
        if self.stdout is None:
            stdout_task = None
        else:
            stdout_task = Task(self._read_close_stdout_stream(), loop)
            tasks.append(stdout_task)
        
        if self.stderr is None:
            stderr_task = None
        else:
            stderr_task = Task(self._read_close_stderr_stream(), loop)
            tasks.append(stderr_task)
        
        if tasks:
            future = WaitTillAll(tasks, loop)
            if (timeout is not None):
                future_or_timeout(future, timeout)
            
            done, pending = await future
            
            if pending:
                # timeout occurred, cancel the read tasks and raise TimeoutExpired.
                for task in pending:
                    task.cancel()
                
                process = self.process
                if process is None:
                    args = None
                else:
                    args = process.args
                
                raise TimeoutExpired(args, timeout)
            
            if (stdin_task is not None):
                stdin_task.get_result()
            
            if (stdout_task is None):
                stdout = None
            else:
                stdout = stdout_task.get_result()
            
            if (stderr_task is None):
                stderr = None
            else:
                stderr = stderr_task.get_result()
        else:
            stdout = None
            stderr = None
        
        await self.wait()
        
        return stdout, stderr
    
    
    def poll(self):
        """
        Returns the subprocess' return code if terminated.
        
        Returns
        -------
        return_code : `None`, `int`
        """
        return self.return_code
    
    
    def pause_writing(self):
        """
        Called when the write transport's buffer goes over the high-water mark.
        
        ``.pause_writing`` is called when the write buffer goes over the high-water mark, and eventually
        ``.resume_writing`` is called when the write buffer size reaches the low-water mark.
        """
        self._paused = True
    
    
    def resume_writing(self):
        """
        Called when the transport's write buffer drains below the low-water mark.
        
        See ``.pause_writing`` for details.
        """
        self._paused = False
        
        drain_waiter = self._drain_waiter
        if drain_waiter is None:
            return
        
        self._drain_waiter = None
        drain_waiter.set_result_if_pending(None)
    
    
    def _do_connection_lost(self, exception):
        """
        Sets ``._connection_lost`` as `True` and set ``._drain_waiter`` result or exception as well.
        
        Parameters
        ----------
        exception : `None`, `BaseException`
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        self._connection_lost = True
        # Wake up the writer if currently paused.
        if not self._paused:
            return
        
        drain_waiter = self._drain_waiter
        if drain_waiter is None:
            return
        
        self._drain_waiter = None
        if exception is None:
            drain_waiter.set_result_if_pending(None)
        else:
            drain_waiter.set_exception_if_pending(exception)
    
    
    async def _drain_helper(self):
        """
        Called when the transport buffer after writing goes over the high limit mark to wait till it is drained.
        
        This method is a coroutine.
        """
        if self._connection_lost:
            raise ConnectionResetError('Connection lost')
        
        if not self._paused:
            return
        
        self._drain_waiter = drain_waiter = Future(self)
        await drain_waiter



__all__ = ('WebSocketServer', )

from functools import partial as partial_func

from ..core import Task, WaitTillAll, skip_poll_cycle
from ..utils import IgnoreCaseMultiValueDictionary

from .websocket_server_protocol import WebSocketServerProtocol


class WebSocketServer:
    """
    Asynchronous websocket server implementation.
    
    Attributes
    ----------
    loop : ``EventThread``
        The event loop to what the websocket server is bound to.
    websockets : `set` of (``WebSocketServerProtocol``, `Any`)
        Active server side asynchronous websocket protocol implementations.
    close_connection_task : `None`, ``Task`` of ``_close``
        Close connection task, what's result is set, when closing of the websocket is done.
        
        Should not be cancelled.
        
        Set, when ``.close`` is called.
    handler : `async-callable`
        An asynchronous callable, what will handle a websocket connection.
        
        Should be given as an `async-callable` accepting `1` parameter the respective asynchronous server side websocket
        protocol implementations.
    server : `None`, ``Server``
        Asynchronous server instance. Set meanwhile the websocket server is running.
    protocol_parameters : `tuple` of `Any`
        WebSocket protocol parameters.
        
        Contains the following elements:
            - `handler` : `async-callable` Same as ``.handler``.
            - `host` : `None`, `str`, `iterable` of (`None`, `str`). To what network interfaces the server be bound.
            - `port` :  `None`, `int`. The port used by the `host`(s).
            - `is_ssl` : `bool`
                Whether the server is secure.
            - `origin` : `None`, `str`. Value of the Origin header.
            - `available_extensions` : `None` or (`list` of `Any`).Available websocket extensions.
                Each websocket extension should have the following `4` attributes / methods:
                - `name`: `str`. The extension's name.
                - `request_params` : `list` of `tuple` (`str`, `str`). Additional header parameters of the extension.
                - `decode` : `callable`. Decoder method, what processes a received websocket frame. Should accept `2`
                    parameters: The respective websocket ``Frame``, and the ˙max_size` as `int`, what describes the
                    maximal size of a received frame. If it is passed, ``PayloadError`` is raised.
                - `encode` : `callable`. Encoder method, what processes the websocket frames to send. Should accept `1`
                    parameter, the respective websocket ``Frame``.
            - `available_subprotocols` : `None` or (`list` of `str`). A list of supported subprotocols in order of
                decreasing preference.
            - `extra_response_headers` : `None` or (``IgnoreCaseMultiValueDictionary``, `dict-like`) of (`str`, `str`) items. Extra
                headers to send with the http response.
            - `request_processor` : `None`, `callable`. An optionally asynchronous callable, what processes the
                initial requests from the potential clients. Should accept the following parameters:
                - `path` : `str`. The requested path.
                - `request_headers` : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`). The request's headers.
                
                The `request_processor` on accepted request should return `None`, otherwise a `tuple` of
                ``AbortHandshake`` parameters.
            - `subprotocol_selector` : `None`, `callable`. User hook to select subprotocols. Should accept the
                following parameters:
                - `parsed_header_subprotocols` : `list` of `str`. The subprotocols supported by the client.
                - `available_subprotocols` : `list` of `str`. The subprotocols supported by the server.
            - `websocket_kwargs` : `dict` of (`str`, `Any`). Extra parameters for creating the websocket protocol.
                
                Can have any of the following items:
                - `close_timeout` : `float`. The maximal duration in seconds what is waited for response after close
                    frame is sent. Defaults to `10.0`.
                - `max_size` : `int`.Max payload size to receive. If a payload exceeds it, ``PayloadError`` is raised.
                    Defaults to `67108864` bytes.
                - `max_queue` : `None`, `int`.
                    Max queue size of ``.messages``. If a new payload is added to a full queue, the oldest element of
                    it is removed. Defaults to `None`.
    """
    __slots__ = ('loop', 'websockets', 'close_connection_task', 'handler', 'server', 'protocol_parameters')
    
    async def __new__(cls, loop, host, port, handler, *, protocol=WebSocketServerProtocol, available_extensions=None,
            extra_response_headers=None, origin=None, available_subprotocols=None, request_processor=None,
            subprotocol_selector=None, websocket_kwargs=None, ssl=None, **server_kwargs):
        """
        Creates a new ``WebSocketServer`` with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the websocket server is bound to.
        host : `None`, `str`, `iterable` of (`None`, `str`)
            To what network interfaces should the server be bound.
        port : `None`, `int`
            The port to use by the `host`(s).
        handler : `async-callable`
            An asynchronous callable, what will handle a websocket connection.
            
            Should be given as an `async-callable` accepting `1` parameter the respective asynchronous server side
            websocket protocol implementations.
        protocol : `Any` = ``WebSocketServerProtocol``, Optional (Keyword only)
            Asynchronous server side websocket protocol implementation.
        available_extensions : `None` or (`list` of `Any`) = `None`, Optional (Keyword only)
            Available websocket extensions.
            
            Each websocket extension should have the following `4` attributes / methods:
            - `name`: `str`. The extension's name.
            - `request_params` : `list` of `tuple` (`str`, `str`). Additional header parameters of the extension.
            - `decode` : `callable`. Decoder method, what processes a received websocket frame. Should accept `2`
                parameters: The respective websocket ``Frame``, and the ˙max_size` as `int`, what decides the
                maximal size of a received frame. If it is passed, ``PayloadError`` is raised.
            - `encode` : `callable`. Encoder method, what processes the websocket frames to send. Should accept `1`
                parameter, the respective websocket ``Frame``.
        extra_response_headers : `None` or (``IgnoreCaseMultiValueDictionary``, `dict-like`) of
                (`str`, `str`) items = `None`, Optional (Keyword only)
            Extra headers to send with the http response.
        origin : `None`, `str` = `None`, Optional (Keyword only)
            Value of the Origin header.
        available_subprotocols : `None` or (`list` of `str`) = `None`, Optional (Keyword only)
            A list of supported subprotocols in order of decreasing preference.
        request_processor : `None`, `callable` = `None`, Optional (Keyword only)
            An optionally asynchronous callable, what processes the initial requests from the potential clients.
            
            Should accept the following parameters:
            - `path` : `str`. The requested path.
            - `request_headers` : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`). The request's headers.
            
            The `request_processor` on accepted request should return `None`, otherwise a `tuple` of
            ``AbortHandshake`` parameters.
        subprotocol_selector `None`, `callable` = `None`, Optional (Keyword only)
            User hook to select subprotocols. Should accept the following parameters:
            - `parsed_header_subprotocols` : `list` of `str`. The subprotocols supported by the client.
            - `available_subprotocols` : `list` of `str`. The subprotocols supported by the server.
        websocket_kwargs : `None`, `dict` of (`str`, `Any`) = `None`, Optional (Keyword only)
            Extra parameters for creating the websocket protocol.
            
            Can have any of the following items:
            - `close_timeout` : `float`. The maximal duration in seconds what is waited for response after close
                frame is sent.
            - `max_size` : `int`.Max payload size to receive. If a payload exceeds it, ``PayloadError`` is raised.
            - `max_queue` : `None`, `int`.
                Max queue size of ``.messages``. If a new payload is added to a full queue, the oldest element of
                it is removed.
        ssl : `None`, ``SSLContext`` = `None`, Optional (Keyword only)
            Whether and what ssl is enabled for the connections.
        **server_kwargs : Keyword parameters
            Additional keyword parameters to create the websocket server with.
        
        Other Parameters
        ----------------
        family : `AddressFamily`, `int`, Optional (Keyword only)
            Can be given either as `socket.AF_INET`, `socket.AF_INET6` to force the socket to use `IPv4`, `IPv6`.
            If not given, then  will be determined from host name.
        backlog : `int`, Optional (Keyword only)
            The maximum number of queued connections passed to `socket.listen()`.
        reuse_address : `bool`, Optional (Keyword only)
            Tells the kernel to reuse a local socket in `TIME_WAIT` state, without waiting for its natural timeout to
            expire. If not specified will automatically be set to True on Unix.
        reuse_port : `bool`, Optional (Keyword only)
            Tells to the kernel to allow this endpoint to be bound to the same port as an other existing endpoint
            already might be bound to.
            
            Not supported on Windows.
        
        Returns
        -------
        self : ``WebSocketServer``
        
        Raises
        ------
        TypeError
            - `extra_response_headers` is not given as `None`, neither as `dict-like`.
            - If `ssl` is not given either as `None`, `ssl.SSLContext`.
            - If `reuse_port` is given as non `bool`.
            - If `reuse_address` is given as non `bool`.
            - If `reuse_port` is given as non `bool`.
            - If `host` is not given as `None`, `str` and neither as `iterable` of `None`, `str`.
        ValueError
            - If `host`, `port` parameter is given, when `socket` is defined as well.
            - If `reuse_port` is given as `True`, but not supported.
            - If neither `host`, `port nor `socket` were given.
            - If `socket` is given, but it's type is not `module_socket.SOCK_STREAM`.
        OsError
            Error while attempting to binding to address.
        """
        if websocket_kwargs is None:
            websocket_kwargs = {}
        
        is_ssl = (ssl is not None)
        
        if available_extensions is None:
            available_extensions = []
        
        if (extra_response_headers is None):
            pass
        elif type(extra_response_headers) is IgnoreCaseMultiValueDictionary:
            pass
        elif hasattr(type(extra_response_headers), 'items'):
            extra_response_headers_local = IgnoreCaseMultiValueDictionary()
            
            for name, value in extra_response_headers.items():
                extra_response_headers_local[name] = value
            
            extra_response_headers = extra_response_headers_local
        else:
            raise TypeError(
                f'`extra_response_headers` can be `None`, `dict-like` with \'.items\' method, got '
                f'{extra_response_headers.__class__.__name__}; {extra_response_headers!r}.'
            )
        
        if (extra_response_headers is not None) and (not extra_response_headers):
            extra_response_headers = None
        
        self = object.__new__(cls)
        self.loop = loop
        self.handler = handler
        self.websockets = set()
        self.close_connection_task = None
        self.server = None
        self.protocol_parameters = (handler, host, port, is_ssl, origin, available_extensions, available_subprotocols,
            extra_response_headers, request_processor, subprotocol_selector, websocket_kwargs)
        
        factory = partial_func(protocol, self,)
        server = await loop.create_server_to(factory, host, port, ssl=ssl, **server_kwargs)
        
        self.server = server
        await server.start()
        
        return self
    
    def register(self, protocol):
        """
        Registers a newly created server side websocket to the websocket server itself.
        
        Parameters
        ----------
        protocol : ``WebSocketServerProtocol``, `Any`
            The connected server side websocket.
        """
        self.websockets.add(protocol)
    
    
    def unregister(self, protocol):
        """
        Unregisters a newly created server side websocket from the websocket server itself.
        
        Parameters
        ----------
        protocol : ``WebSocketServerProtocol``, `Any`
            The disconnected server side websocket.
        """
        self.websockets.discard(protocol)
    
    
    def is_serving(self):
        """
        Returns whether the websocket server is serving.
        
        Returns
        -------
        is_serving : `bool`
        """
        server = self.server
        if server is None:
            return False
        
        if server.sockets is None:
            return False
        
        return True
    
    
    def close(self):
        """
        Closes the websocket server. Returns a closing task, what can be awaited.
        
        Returns
        -------
        close_connection_task : ``Task`` of ``_close``
            Close connection task, what's result is set, when closing of the websocket is done.
            
            Should not be cancelled.
        """
        close_connection_task = self.close_connection_task
        if close_connection_task is None:
            close_connection_task = Task(self._close(), self.loop)
            self.close_connection_task = close_connection_task
        
        return close_connection_task
    
    
    async def _close(self):
        """
        Closes the websocket server. If the websocket task is already closed does nothing.
        
        This method is a coroutine.
        """
        server = self.server
        if server is None:
            return
        
        server.close()
        await server.wait_closed()
        
        loop = self.loop
        
        # Skip 1 full loop
        await skip_poll_cycle(loop)
        
        websockets = self.websockets
        if websockets:
            tasks = []
            for websocket in websockets:
                tasks.append(websocket.close(1001))
            
            future = WaitTillAll(tasks, loop)
            tasks = None
            await future
            
        if websockets:
            tasks = []
            for websocket in websockets:
                task = websocket.handler_task
                if task is None:
                    continue
                
                tasks.append(task)
            
            task = None
            if tasks:
                future = WaitTillAll(tasks, loop)
                tasks = None
                await future



__all__ = ()

# Test for pypy bug:
# https://foss.heptapod.net/pypy/pypy/issues/3239
class dummy_init_tester:
    def __new__(cls, value):
        return object.__new__(cls)
    __init__ = object.__init__

try:
    dummy_init_tester(None)
except TypeError:
    NEEDS_DUMMY_INIT = True
else:
    NEEDS_DUMMY_INIT = False

del dummy_init_tester



__all__ = (
    'IS_UNIX', 'any_to_any', 'change_on_switch', 'get_short_executable', 'is_hashable', 'is_iterable',
    'list_difference', 'relative_index', 'un_map_pack', 'where'
)

import sys
from os import sep as PATH_SEPARATOR

from .docs import has_docs


IS_UNIX = (sys.platform != 'win32')

@has_docs
def any_to_any(container1, container2):
    """
    Returns whether any value of `container1` is in `container2` as well.
    
    Parameters
    ----------
    container1 : `iterable-container`
        Any iterable container.
    container2 : `iterable-container`
        Any iterable container.

    Returns
    -------
    contains : `bool`
    """
    for value in container1:
        if value in container2:
            return True
    
    return False


@has_docs
def get_short_executable():
    """
    Gets short executable name.
    
    Returns
    -------
    executable : `str`
    """
    executable = sys.executable
    
    index = executable.rfind(PATH_SEPARATOR)
    if index != -1:
        executable = executable[index + 1:]
    
    if (not IS_UNIX) and executable.endswith('.exe'):
        executable = executable[:-len('.exe')]
    
    return executable


@has_docs
def where(container, key):
    """
    Returns the first element from the given container on what `key` returns `True`.
    
    Parameters
    ----------
    container : `iterable-container`
        Any iterable container.
    key : `function`
        Function used to determine whether an element of `container` meets our expectations.
    
    Returns
    -------
    value : `Any`
        An element of `container`.
    
    Raises
    ------
    LookupError
        On non of the elements was `true` returned by they `key.
    """
    for value in container:
        if key(value):
            break
    else:
        raise LookupError(key)
    
    return value


@has_docs
def relative_index(list_, value):
    """
    Returns on which the given `value` would be inserted into the given list.
    
    Parameters
    ----------
    list_ : `list` of `Any`
        The list o which value would be inserted.
    value : `Any`
        The value what would be inserted.
    
    Returns
    -------
    relative_index : `int`
    """
    bot = 0
    top = len(list_)
    
    while True:
        if bot < top:
            half = (bot + top) >> 1
            if list_[half] < value:
                bot = half + 1
            else:
                top = half
            continue
        return bot


@has_docs
def change_on_switch(list_, value, new_position, key=None):
    """
    Calculates the changes if the given `value` would be moved to an another position.
    
    Parameters
    ----------
    list_ : `list` of `Any`
        The list on what the changes will be calculated.
    value : `Any`
        The object, what would be moved.
    new_position : `int`
        The new position of the value.
    key : `None`, `callable`
        A special callable what would be used to used to build each element of the result.
    
    Returns
    -------
    result : `list` of (`tuple` (`int`, `Any`)) or `callable` returns
        The changed positions.
    
    Raises
    ------
    ValueError
        The given `value` is not in the list.
    """
    ln = len(list_)
    
    old_position = relative_index(list_, value)
    if (old_position == ln) or (list_[old_position] != value):
        raise ValueError(f'{value!r} is not in the {list_.__class__.__name__}.')
    
    if new_position >= ln:
        new_position = ln - 1
    elif new_position < 0:
        new_position = 0
    
    result = []
    if new_position == old_position:
        return result
    
    if new_position < old_position:
        index = new_position
        limit = old_position
        change = +1
    
    else:
        index = old_position + 1
        limit = new_position + 1
        change = -1
    
    while True:
        actual = list_[index]
        index += 1
        
        position = index + change
        if key is None:
            element = (actual, position)
        else:
            element = key(actual, position)
        
        result.append(element)
        
        if index == limit:
            break
        
        continue
    
    if key is None:
        element = (value, new_position)
    else:
        element = key(value, new_position)
    
    if change > 0:
        result.index(0, element)
    else:
        result.append(element)
    
    return result



@has_docs
def list_difference(list1, list2):
    """
    Returns the difference of the two given lists.
    
    Parameters
    ----------
    list1 : `None`, `list`, `set`
        The first list, what's inclusive elements should go into the return's zero-th element.
    list2 : `None`, `list`, `set`
        The second list, what's inclusive elements should go into the return's one-th element.
    
    Returns
    -------
    difference : `tuple` (`list` of `Any`, `list` of `Any`)
        A tuple containing the inclusive element of the given parameters.
    
    Notes
    -----
    `list1` and `list2` should be given as sorted lists, but `None` and `set`-s are also accepted.
    """
    difference = ([], [])
    
    if list1 is None:
        if list2 is None:
            return difference
        else:
            difference[1].extend(list2)
            return difference
    else:
        if list2 is None:
            difference[0].extend(list1)
            return difference
    
    if isinstance(list1, set):
        list1 = sorted(list1)
    if isinstance(list2, set):
        list2 = sorted(list2)
        
    ln1 = len(list1)
    ln2 = len(list2)
    index1 = 0
    index2 = 0

    #some quality python here again *cough*
    while True:
        if index1 == ln1:
            while True:
                if index2 == ln2:
                    break
                value2 = list2[index2]
                difference[1].append(value2)
                index2 += 1

            break
        if index2 == ln2:
            while True:
                if index1 == ln1:
                    break
                value1 = list1[index1]
                difference[0].append(value1)
                index1 += 1

            break
        
        value1 = list1[index1]
        value2 = list2[index2]
        if value1 < value2:
            difference[0].append(value1)
            index1 += 1
            continue
        if value1 > value2:
            difference[1].append(value2)
            index2 += 1
            continue
        if value1 != value2:
            difference[0].append(value1)
            difference[1].append(value2)
        index1 += 1
        index2 += 1
    
    return difference


@has_docs
class un_map_pack:
    """
    Helper class for un-map-packing generators.
    
    Attributes
    ----------
    generator : ``GeneratorType``
        The generator to unpack.
    next_value : `Any`
        The next key to return on get-item.
    """
    def __init__(self, generator):
        """
        Creates a new generator into mapping unpacker.
        
        Parameters
        ----------
        generator : ``GeneratorType``
            A generator to un-map-pack.
        """
        self.generator = generator
        self.next_value = None
    
    @has_docs
    def keys(self):
        """
        Calling ``.keys`` returns itself.
        
        Returns
        -------
        self: ``un_map_pack``
        """
        return self
    
    @has_docs
    def __iter__(self):
        """Iterating an un map packer returns itself."""
        return self
    
    @has_docs
    def __next__(self):
        """Gets the next key of the un-map-packer"""
        key, value = next(self.generator)
        self.next_value = value
        return key
    
    @has_docs
    def __getitem__(self, key):
        """Gets the next value of the un-map-packer."""
        return self.next_value


@has_docs
def is_iterable(object_):
    """
    Returns whether the object is iterable.
    
    Parameters
    ----------
    object_ : `Any`
        The object to check.
    
    Returns
    -------
    is_iterable : `bool`
    """
    return hasattr(type(object_), '__iter__')


@has_docs
def is_hashable(object_):
    """
    Returns whether the object is hashable.
    
    Parameters
    ----------
    object_ : `Any`
        The object to check.
    
    Returns
    -------
    is_hashable : `bool`
    """
    try:
        hasher_function = getattr(type(object_), '__hash__')
    except AttributeError:
        return False
    
    try:
        hasher_function(object_)
    except (TypeError, NotImplementedError):
        return False
    
    return True



from .abstract import *
from .extra_info import *
from .protocol import *
from .ssl_pipe import *
from .ssl_transport_layer import *
from .transport_layer import *
from .unix_pipe_transport_layer import *

__all__ = (
    *abstract.__all__,
    *extra_info.__all__,
    *protocol.__all__,
    *ssl_pipe.__all__,
    *ssl_transport_layer.__all__,
    *transport_layer.__all__,
    *unix_pipe_transport_layer.__all__,
)



__all__ = ()

from .compressors import ZLIB_COMPRESSOR, ZLIB_MAX_WBITS


WRITE_CHUNK_LIMIT = 65536

class HTTPStreamWriter:
    """
    Http writer used by ``ClientRequest``.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether ``.write_eof`` was called.
    size : `int`
        The amount of written data in bytes. If reaches a limit, drain lock is awaited.
    chunked : `bool`
        Whether the http message's content is chunked.
    compressor : `None`, `ZLIB_COMPRESSOR`, `BROTLI_COMPRESSOR`
        Decompressor used to compress the sent data. Defaults to `None` if no compression is given.
    protocol : `ProtocolBase`
        Asynchronous transport implementation.
    transport : `None`, `AbstractTransportLayerBase`
        Asynchronous transport implementation. Set as `None` if at eof.
    """
    __slots__ = ('_at_eof', 'size', 'chunked', 'compressor', 'protocol', 'transport', )
    
    def __init__(self, protocol, compression, chunked):
        """
        Creates a new ``HTTPStreamWriter`` with the given parameter.
        
        Parameters
        ----------
        protocol : `Any`
            Asynchronous transport implementation.
        compression : `None`, `str`
            The compression's type to encode the written content with.
        chunked : `bool`
            Whether the given data should be written with chunking.
        """
        if (compression is None):
            compressor = None
        elif compression == 'gzip':
            compressor = ZLIB_COMPRESSOR(wbits=16 + ZLIB_MAX_WBITS)
        elif compression == 'deflate':
            compressor = ZLIB_COMPRESSOR(wbits=ZLIB_MAX_WBITS)
        else:
            compressor = None
        
        self.compressor = compressor
        
        self.protocol = protocol
        self.transport = protocol.get_transport()
        
        self.chunked = chunked
        self.size = 0
        
        self._at_eof = False
    
    def _write(self, chunk):
        """
        Writes the given chunk of data to the writer's ``.transport``.
        
        Parameters
        ----------
        chunk : `bytes-like`
            The data to write.
        
        Raises
        ------
        ConnectionResetError
            Cannot write to closing transport.
        """
        size = len(chunk)
        self.size += size
        
        transport = self.transport
        if (transport is None) or transport.is_closing():
            raise ConnectionResetError('Cannot write to closing transport.')
        
        transport.write(chunk)
    
    async def write(self, chunk):
        """
        Writes the given chunk of data to the writer's ``.transport``.
        
        Compresses and "chunks" the `chunk` if applicable. If after writing the respective buffer limit is reached,
        waits for drain lock.
        
        This method is a coroutine.
        
        Parameters
        ----------
        chunk : `bytes-like`
            The data to write.
        
        Raises
        ------
        ConnectionResetError
            Cannot write to closing transport.
        """
        compressor = self.compressor
        if (compressor is not None):
            chunk = compressor.compress(chunk)
        
        if not chunk:
            return
        
        if self.chunked:
            chunk = b''.join([len(chunk).__format__('x').encode('ascii'), b'\r\n', chunk, b'\r\n'])
        
        self._write(chunk)
        
        if self.size > WRITE_CHUNK_LIMIT:
            self.size = 0
            protocol = self.protocol
            if protocol.get_transport() is not None:
                await protocol._drain_helper()
    
    
    async def write_eof(self, chunk=b''):
        """
        Write end of stream to the writer's ``.transport`` and marks it as it is at eof.
        
        If the writer is already at eof, does nothing.
        
        This method is a coroutine.
        
        Parameters
        ----------
        chunk : `bytes-like` = `b''`, Optional
            The data to write.
        """
        if self._at_eof:
            return
        
        compressor = self.compressor
        if compressor is None:
            if self.chunked:
                if chunk:
                    chunk = b''.join([len(chunk).__format__('x').encode('ascii'), b'\r\n', chunk, b'\r\n0\r\n\r\n'])
                else:
                    chunk = b'0\r\n\r\n'
        else:
            if chunk:
                chunk = compressor.compress(chunk)
                chunk = chunk + compressor.flush()
            else:
                chunk = compressor.flush()
            
            if chunk and self.chunked:
                chunk = b''.join([len(chunk).__format__('x').encode('ascii'), b'\r\n', chunk, b'\r\n0\r\n\r\n'])
        
        if chunk:
            self._write(chunk)
        
        protocol = self.protocol
        if protocol.get_transport() is not None:
            await protocol._drain_helper()
        
        self._at_eof = True
        self.transport = None
    
    
    async def drain(self):
        """
        Flushes the write buffer.
        """
        protocol = self.protocol
        if (protocol._transport is not None):
            await protocol._drain_helper()



from .subprocess import *
from .subprocess_protocols import *
from .subprocess_writer import *

__all__ = (
    *subprocess.__all__,
    *subprocess_protocols.__all__,
    *subprocess_writer.__all__,
)
